from __future__ import annotations

import argparse

from ramalama.plugins.interface import RuntimePlugin
from ramalama.plugins.registry import PluginRegistry

# Registry for runtime plugins (entry point group: "ramalama.runtimes.v1alpha")
# Future plugin types register their own PluginRegistry with a distinct group name, e.g.:
#   CHAT_PLUGIN_REGISTRY = PluginRegistry("ramalama.chat_providers", ChatPlugin)
_RUNTIME_REGISTRY: PluginRegistry[RuntimePlugin] = PluginRegistry("ramalama.runtimes.v1alpha", RuntimePlugin)  # type: ignore[type-abstract]


def get_all_runtimes() -> dict[str, RuntimePlugin]:
    return _RUNTIME_REGISTRY.load()


def get_runtime(name: str) -> RuntimePlugin:
    plugin = _RUNTIME_REGISTRY.get(name)
    if plugin is None:
        raise ValueError(f"Unknown runtime: '{name}'")
    return plugin


def assemble_command(cli_args: argparse.Namespace) -> list[str]:
    runtime = str(cli_args.runtime)
    command = str(cli_args.subcommand)
    return get_runtime(runtime).handle_subcommand(command, cli_args)



from __future__ import annotations

import json
import struct

import ramalama.console as console
from ramalama.model_inspect.error import ParseError
from ramalama.model_inspect.safetensor_info import SafetensorModelInfo

# Based on safetensor format description:
# https://github.com/huggingface/safetensors?tab=readme-ov-file#format


class SafetensorInfoParser:
    @staticmethod
    def is_model_safetensor(model_name: str) -> bool:

        # There is no magic number or something similar, so we only rely on the naming of the file here
        return model_name.endswith(".safetensor") or model_name.endswith(".safetensors")

    @staticmethod
    def parse(model_name: str, model_registry: str, model_path: str) -> SafetensorModelInfo:
        try:
            with open(model_path, "rb") as model_file:
                prefix = '<'
                typestring = f"{prefix}Q"

                header_size = struct.unpack(typestring, model_file.read(8))[0]
                header = json.loads(model_file.read(header_size))

                return SafetensorModelInfo(model_name, model_registry, model_path, header)

        except Exception as ex:
            msg = f"Failed to parse safetensor model '{model_path}': {ex}"
            console.warning(msg)
            raise ParseError(msg)



from __future__ import annotations

import base64
import mimetypes

from ramalama.file_loaders.file_types.base import BaseFileLoader


class BasicImageFileLoader(BaseFileLoader):
    @staticmethod
    def file_extensions() -> set[str]:
        return {
            ".jpg",
            ".jpeg",
            ".png",
            ".gif",
            ".bmp",
            ".tiff",
            ".tif",
            ".webp",
            ".ico",
        }

    @staticmethod
    def load(file: str) -> str:
        """
        Load the content of the text file.
        """

        mime_type, _ = mimetypes.guess_type(file)
        with open(file, "rb") as f:
            data = base64.b64encode(f.read()).decode("utf-8")

        return f"data:{mime_type};base64,{data}"



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Optional, TypedDict
from urllib import error as urllib_error
from urllib import request as urllib_request

from ramalama.chat_utils import ChatMessageType
from ramalama.config import ActiveConfig


class RequiredChatRequestOptionsDict(TypedDict):
    stream: bool


class ChatRequestOptionsDict(RequiredChatRequestOptionsDict, total=False):
    model: str
    temperature: float
    max_tokens: int


@dataclass
class ChatRequestOptions:
    """Normalized knobs for building a chat completion request."""

    model: Optional[str]
    stream: bool = True
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    extra: Optional[dict[str, Any]] = None

    def to_dict(self) -> ChatRequestOptionsDict:
        result: ChatRequestOptionsDict = {'stream': self.stream}

        if self.temperature is not None:
            result['temperature'] = self.temperature
        if self.max_tokens is not None:
            result['max_tokens'] = self.max_tokens
        if self.model is not None:
            result['model'] = self.model

        if self.extra is not None:
            # Can't easily type this without access to open TypedDicts
            # https://typing.python.org/en/latest/spec/glossary.html#term-closed
            result |= self.extra  # type: ignore

        return result


@dataclass
class ChatStreamEvent:
    """A provider-agnostic representation of a streamed delta."""

    text: Optional[str] = None
    raw: Optional[dict[str, Any]] = None
    done: bool = False


class ChatProviderError(Exception):
    """Raised when a provider request fails or returns an invalid payload."""

    def __init__(self, message: str, *, status_code: Optional[int] = None, payload: Optional[Any] = None):
        super().__init__(message)
        self.status_code = status_code
        self.payload = payload


class ChatProvider(ABC):
    """Abstract base class for hosted chat providers."""

    provider: str = "base"
    default_path: str

    def __init__(
        self,
        base_url: str,
        api_key: Optional[str] = None,
        default_headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        if api_key is None:
            api_key = ActiveConfig().api_key

        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self._default_headers: dict[str, str] = dict(default_headers or {})

    def build_url(self, path: Optional[str] = None) -> str:
        rel = path or self.default_path
        if not rel.startswith("/"):
            rel = f"/{rel}"
        return f"{self.base_url}{rel}"

    def prepare_headers(
        self,
        *,
        include_auth: bool = True,
        extra: Optional[dict[str, str]] = None,
        options: Optional[ChatRequestOptions] = None,
    ) -> dict[str, str]:
        headers: dict[str, str] = {
            "Content-Type": "application/json",
            **self._default_headers,
            **self.provider_headers(options),
        }

        if include_auth:
            headers.update(self.auth_headers())
        if extra:
            headers.update(extra)
        return headers

    def auth_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.api_key}"} if self.api_key else {}

    def serialize_payload(self, payload: Mapping[str, Any]) -> bytes:
        return json.dumps(payload).encode("utf-8")

    def create_request(
        self, messages: Sequence[ChatMessageType], options: ChatRequestOptions
    ) -> urllib_request.Request:
        payload = self.build_payload(messages, options)
        headers = self.prepare_headers(options=options, extra=self.additional_request_headers(options))
        body = self.serialize_payload(payload)
        return urllib_request.Request(
            self.build_url(self.resolve_request_path(options)),
            data=body,
            headers=headers,
            method="POST",
        )

    # ------------------------------------------------------------------
    # Provider customization points
    # ------------------------------------------------------------------
    def provider_headers(self, options: Optional[ChatRequestOptions] = None) -> dict[str, str]:
        return {}

    def additional_request_headers(self, options: Optional[ChatRequestOptions] = None) -> dict[str, str]:
        return {}

    def resolve_request_path(self, options: Optional[ChatRequestOptions] = None) -> str:
        return self.default_path

    @abstractmethod
    def build_payload(self, messages: Sequence[ChatMessageType], options: ChatRequestOptions) -> Mapping[str, Any]:
        """Return the provider-specific payload."""

    @abstractmethod
    def parse_stream_chunk(self, chunk: bytes) -> Iterable[ChatStreamEvent]:
        """Yield zero or more events parsed from a streamed response chunk."""

    # ------------------------------------------------------------------
    # Error handling
    # ------------------------------------------------------------------
    def raise_for_status(self, status_code: int, payload: Optional[Any] = None) -> None:
        if status_code >= 400:
            if isinstance(payload, dict) and "error" in payload:
                err = payload["error"]
                message = str(err.get("message") or err.get("type") or err) if isinstance(err, dict) else str(err)
            else:
                message = "chat request failed"

            raise ChatProviderError(message, status_code=status_code, payload=payload)

    # ------------------------------------------------------------------
    # Non-streamed helpers
    # ------------------------------------------------------------------
    def parse_response_body(self, body: bytes) -> Any:
        if not body:
            return None
        return json.loads(body.decode("utf-8"))

    def list_models(self) -> list[str]:
        """Return available model identifiers exposed by the provider."""

        request = urllib_request.Request(
            self.build_url("/models"),
            headers=self.prepare_headers(include_auth=True),
            method="GET",
        )
        try:
            with urllib_request.urlopen(request) as response:
                payload = self.parse_response_body(response.read())
        except urllib_error.HTTPError as exc:
            if exc.code in (401, 403):
                message = (
                    f"Could not authenticate with {self.provider}."
                    " The provided API key was either missing or invalid.\n"
                    f"Set RAMALAMA_API_KEY or ramalama.provider.<provider_name>.api_key."
                )
                try:
                    payload = self.parse_response_body(exc.read())
                except Exception:
                    payload = {}

                if details := payload.get("error", {}).get("message", None):
                    message = f"{message}\n\n{details}"

                raise ChatProviderError(message, status_code=exc.code) from exc
            raise

        if not isinstance(payload, Mapping):
            raise ChatProviderError("Invalid model list payload", payload=payload)

        data = payload.get("data")
        if not isinstance(data, list):
            raise ChatProviderError("Invalid model list payload", payload=payload)

        models: list[str] = []
        for entry in data:
            if isinstance(entry, Mapping) and (model_id := entry.get("id")):
                models.append(str(model_id))

        return models


__all__ = [
    "ChatProvider",
    "ChatProviderError",
    "ChatRequestOptions",
    "ChatStreamEvent",
]



#!/usr/bin/env python3
from __future__ import annotations

import json
import logging
import time
import urllib.error
import urllib.request
from typing import Any, Dict, List, Optional

from ramalama.config import ActiveConfig
from ramalama.logger import logger
from ramalama.mcp.mcp_client import PureMCPClient


class LLMAgent:
    """An LLM-powered agent that can make multiple tool calls to accomplish tasks."""

    def __init__(
        self,
        clients: List[PureMCPClient],
        llm_base_url: str = "http://localhost:8080",
        model: Optional[str] = None,
        args=None,
    ):
        config_level = ActiveConfig().log_level or logging.INFO
        if logging.getLogger().handlers:
            logging.getLogger().setLevel(config_level)
        else:
            logging.basicConfig(level=config_level)
        self.clients = clients if isinstance(clients, list) else [clients]
        self.llm_base_url = llm_base_url.rstrip('/')
        self.model = model
        self.args = args
        self.available_tools: List[Dict[str, Any]] = []
        self.tool_to_client: Dict[str, PureMCPClient] = {}
        self._stream_callback = None

    def initialize(self):
        """Initialize the agent and get available tools from all clients."""
        all_init_results = []
        self.available_tools = []
        self.tool_to_client = {}

        for i, client in enumerate(self.clients):
            try:
                init_result = client.initialize()
                all_init_results.append(init_result)

                server_name = init_result['result']['serverInfo']['name']
                tools_result = client.list_tools()
                server_tools = tools_result['result']['tools']

                for tool in server_tools:
                    tool_name = tool['name']
                    if tool_name in self.tool_to_client:
                        original_name = tool_name
                        tool_name = f"{server_name}_{original_name}"
                        tool['name'] = tool_name

                    tool['server'] = server_name
                    self.tool_to_client[tool_name] = client
                    self.available_tools.append(tool)

            except Exception as e:
                logging.error("Failed to initialize client %s: %s", i, e, exc_info=True)

        if not self.available_tools:
            raise RuntimeError("No tools available from any server")

        return all_init_results, self.available_tools

    def print_tools(self):
        print("\nAvailable tools:\n")
        for i, tool in enumerate(self.available_tools, 1):
            name = tool.get("name", "unknown")
            description = tool.get("description", "").split(" (from")[0]
            logger.debug(f"{name}: {description}")

            input_props = tool.get("inputSchema", {}).get("properties", {})
            if input_props:
                inputs = [f"{key} ({prop.get('type', 'unknown')})" for key, prop in input_props.items()]
                inputs_str = ", ".join(inputs)
            else:
                inputs_str = "none"

            print(f"  {i}. {name}")
            print(f"     Inputs: {inputs_str}\n")

    def should_use_tools(self, content: str, conversation_history: Optional[list[dict[str, str]]] = None) -> bool:
        """Determine if the request should be handled by tools using LLM."""
        tools_context = "Available tools:\n"
        for i, tool in enumerate(self.available_tools, 1):
            name = tool.get("name", "unknown")
            description = tool.get("description", "")
            tools_context += f"{i}. {name}: {description}\n"

        context_info = ""
        if conversation_history:
            context_info = "\nRecent conversation:\n"
            for msg in conversation_history[-3:]:
                role = msg['role'].capitalize()
                preview = msg['content'][:150] + "..." if len(msg['content']) > 150 else msg['content']
                context_info += f"{role}: {preview}\n"

        messages = [
            {
                "role": "system",
                "content": (
                    "You are an intelligent assistant that determines whether a user's "
                    "request should use the available tools.\n\n"
                    "Answer ONLY \"YES\" if tools should be used or \"NO\" otherwise."
                ),
            },
            {
                "role": "user",
                "content": (
                    f"""User request: {content}
{context_info}
{tools_context}

Should this request use the available tools?"""
                ),
            },
        ]
        response = self._call_llm(messages)
        return response is not None and response.upper().strip() == "YES"

    def _call_llm(self, messages: List[Dict[str, str]], console_stream: bool = False) -> Optional[str]:
        """
        Call the LLM with the given messages.
        """
        request_data = {"messages": messages, "stream": True}
        if self.model is not None:
            request_data["model"] = self.model
        data = json.dumps(request_data).encode("utf-8")

        headers = {"Content-Type": "application/json"}

        # Add API key if available
        if self.args and getattr(self.args, "api_key", None):
            headers["Authorization"] = f"Bearer {self.args.api_key}"
        request = urllib.request.Request(
            f"{self.llm_base_url}/chat/completions",
            data=data,
            headers=headers,
            method="POST",
        )

        if not console_stream:
            content = ""
            # Retry logic similar to regular chat system
            max_retries = 10
            retry_delay = 0.5

            for attempt in range(max_retries):
                try:
                    with urllib.request.urlopen(request, timeout=30) as response:
                        for raw_line in response:
                            line = raw_line.decode("utf-8").strip()
                            if line.startswith("data: "):
                                payload = line[6:]
                                if payload.strip() == "[DONE]":
                                    break
                                try:
                                    event = json.loads(payload)
                                    if "choices" in event and event["choices"]:
                                        delta = event["choices"][0].get("delta", {})
                                        if "content" in delta and delta["content"] is not None:
                                            content += delta["content"]
                                except json.JSONDecodeError:
                                    logging.warning("Malformed SSE line: %s", payload)
                        return content.strip()
                except Exception as e:
                    if attempt == max_retries - 1:
                        # Last attempt failed, log error and return empty
                        logging.error("LLM call failed after %d attempts: %s", max_retries, e)
                        return ""
                    else:
                        # Retry after delay
                        logging.debug("LLM call failed (attempt %d/%d), retrying: %s", attempt + 1, max_retries, e)
                        time.sleep(retry_delay)
                        retry_delay *= 2  # Exponential backoff
            return ""
        elif console_stream:
            try:
                with urllib.request.urlopen(request, timeout=30) as response:
                    for raw_line in response:
                        line = raw_line.decode("utf-8").strip()
                        if line.startswith("data: "):
                            payload = line[6:]
                            if payload.strip() == "[DONE]":
                                break
                            try:
                                event = json.loads(payload)
                                if "choices" in event and event["choices"]:
                                    delta = event["choices"][0].get("delta", {})
                                    if "content" in delta and delta["content"] is not None:
                                        if callable(self._stream_callback):
                                            self._stream_callback(delta["content"])
                                        else:
                                            print(delta["content"], end="", flush=True)
                            except json.JSONDecodeError:
                                logging.warning("Malformed SSE line: %s", payload)
                return None
            except Exception as e:
                logging.error("LLM streaming call failed: %s", e, exc_info=True)
                return None
        else:
            raise ValueError(f"Unknown mode: {console_stream}")

    def _get_tool_arguments_manual(self, tool: dict) -> dict:
        """Prompt user based on inputSchema."""
        args: dict[Any, Any] = {}
        input_props = tool.get("inputSchema", {}).get("properties", {})
        required_fields = tool.get("inputSchema", {}).get("required", [])

        for name, info in input_props.items():
            default = info.get("default")
            param_type = info.get("type", "string")
            prompt_str = f"Enter value for {name} ({param_type})"
            if default is not None:
                prompt_str += f" (default={default})"
            prompt_str += ": "

            while True:
                value = input(prompt_str).strip()
                if not value and default is not None:
                    value = default
                if value or name not in required_fields:
                    # Convert value based on type
                    try:
                        if param_type == "integer":
                            args[name] = int(value) if value else None
                        elif param_type == "number":
                            args[name] = float(value) if value else None
                        elif param_type == "boolean":
                            args[name] = value.lower() in ("true", "1", "yes", "on") if value else None
                        else:
                            args[name] = value
                    except ValueError:
                        print(f"Invalid {param_type} value for {name}. Please try again.")
                        continue
                    break
                print(f"{name} is required!")

        return args

    def _get_tool_arguments_auto(self, tool: dict, task: str, stream: bool = False) -> dict:
        """
        Use LLM to infer arguments automatically.
        If stream=True, stream the LLM output to console.
        """
        tool_inputs = tool.get("inputSchema", {}).get("properties", {})
        if not tool_inputs:
            return {}

        # Build argument descriptions
        arg_descriptions = []
        for name, info in tool_inputs.items():
            param_type = info.get('type', 'string')
            description = info.get('description', '')
            arg_descriptions.append(f"- {name} ({param_type}): {description}")

        prompt = f"""Task: {task}
Tool: {tool['name']}
Arguments needed:
{chr(10).join(arg_descriptions)}

Generate ONLY a JSON object with the arguments. Extract values from the task."""

        messages = [
            {
                "role": "system",
                "content": (
                    "You are a JSON generator. Respond ONLY with valid JSON. "
                    "No explanations, no markdown, just the JSON object."
                ),
            },
            {"role": "user", "content": prompt},
        ]

        if stream:
            self._call_llm(messages, console_stream=True)
            return {}
        else:
            response = self._call_llm(messages, console_stream=False) or ""
            try:
                # Extract JSON from markdown code blocks if present
                json_str = response.strip()
                if json_str.startswith("```json") and json_str.endswith("```"):
                    json_str = json_str[7:-3].strip()
                elif json_str.startswith("```") and json_str.endswith("```"):
                    json_str = json_str[3:-3].strip()

                args = json.loads(json_str)
                # Convert types based on schema
                converted_args: dict[Any, Any] = {}
                for k, v in args.items():
                    if k in tool_inputs:
                        param_type = tool_inputs[k].get('type', 'string')
                        try:
                            if param_type == "integer" and v is not None:
                                converted_args[k] = int(v)
                            elif param_type == "number" and v is not None:
                                converted_args[k] = float(v)
                            elif param_type == "boolean" and v is not None:
                                converted_args[k] = bool(v)
                            else:
                                converted_args[k] = v
                        except (ValueError, TypeError):
                            logging.warning("Failed to convert argument %s to %s: %s", k, param_type, v)
                            converted_args[k] = v
                return converted_args
            except (json.JSONDecodeError, TypeError):
                logging.warning("LLM failed to produce valid JSON for tool arguments: %s", response)
                return {}

    def execute_task(self, task: str, manual: bool = False, stream: bool = False) -> Optional[str]:
        """Execute one or more relevant tools for a task and combine results."""
        if not self.available_tools:
            return "No tools available."

        selected_tools = self._select_tools(task)
        if not selected_tools:
            return "No relevant tools found for this task."

        print(f"Selected tools: {', '.join([t['name'] for t in selected_tools])}")

        tool_outputs = []

        for tool in selected_tools:
            try:
                client = self.tool_to_client[tool["name"]]

                # Decide argument source based on manual/auto
                if manual:
                    arguments = self._get_tool_arguments_manual(tool)
                else:
                    arguments = self._get_tool_arguments_auto(tool, task)

                # Call the tool with the arguments
                result = client.call_tool(tool["name"], arguments)

                if "error" in result:
                    tool_outputs.append(f"{tool['name']} error: {result['error']['message']}")
                elif result.get("result", {}).get("isError"):
                    tool_outputs.append(f"{tool['name']} execution failed.")
                else:
                    # Safely extract text content
                    content_list = result.get("result", {}).get("content", [])
                    if content_list and isinstance(content_list, list):
                        text = "\n".join(c.get("text", "") for c in content_list)
                    else:
                        text = str(result.get("result", {}))
                    tool_outputs.append(f"{tool['name']} result:\n{text}")

            except Exception as e:
                tool_outputs.append(f"{tool['name']} exception: {e}")

        combined_output = "\n\n".join(tool_outputs)
        return self._result(task, combined_output, stream)

    def execute_specific_tool(self, task: str, tool_name: str, manual: bool = False) -> Optional[str]:
        """Execute a specific tool by name."""
        if not self.available_tools:
            return "No tools available."

        tool = next((t for t in self.available_tools if t["name"] == tool_name), None)
        if not tool:
            return f"Tool '{tool_name}' not found."

        try:
            client = self.tool_to_client[tool["name"]]

            if manual:
                arguments = self._get_tool_arguments_manual(tool)
            else:
                arguments = self._get_tool_arguments_auto(tool, task)

            result = client.call_tool(tool["name"], arguments)

            if "error" in result:
                return f"Error: {result['error']['message']}"
            elif result.get("result", {}).get("isError"):
                return "Tool execution failed."
            else:
                content_list = result.get("result", {}).get("content", [])
                if content_list and isinstance(content_list, list):
                    text = "\n".join(c.get("text", "") for c in content_list)
                else:
                    text = str(result.get("result", {}))
                return self._result(task, text) if task.strip() else text

        except Exception as e:
            return f"Error executing tool: {e}"

    def _select_tools(self, task: str) -> List[Dict[str, Any]]:
        """Select one or more tools that should be used for the task."""
        if not self.available_tools:
            return []

        if len(self.available_tools) == 1:
            return [self.available_tools[0]]

        tools_context = "Available tools:\n"
        for i, tool in enumerate(self.available_tools, 1):
            name = tool.get("name", "unknown")
            description = tool.get("description", "")
            tools_context += f"{i}. {name}: {description}\n"

        messages = [
            {
                "role": "system",
                "content": (
                    "You are a helpful assistant that selects ALL relevant tools for a given task. "
                    "Respond ONLY with a comma-separated list of tool names. "
                    "If none are useful, respond with NONE."
                ),
            },
            {"role": "user", "content": f"Task: {task}\n\n{tools_context}"},
        ]

        response = self._call_llm(messages)
        if response:
            response = response.strip()
        if not response or response.upper() == "NONE":
            return []

        chosen = [name.strip().lower() for name in response.split(",")]
        return [t for t in self.available_tools if t["name"].lower() in chosen]

    def _result(self, task: str, content: str, stream: bool = False) -> Optional[str]:
        """Format the tool result into a user-friendly response."""
        messages = [
            {
                "role": "system",
                "content": (
                    "You are a helpful assistant that formats and presents information clearly. "
                    "Understand the user request, analyze the raw tool output, and provide a "
                    "clear, well-structured answer."
                ),
            },
            {"role": "user", "content": f"Request: {task}\n\nRaw tool output:\n{content}"},
        ]
        result = self._call_llm(messages, console_stream=stream)
        if stream and result is None:
            print("")
        return result

    def close(self):
        """Shutdown all MCP clients."""
        for client in self.clients:
            try:
                client.close()
            except Exception as e:
                logging.debug("Error closing MCP client %s", e)



from __future__ import annotations

import json
import os
import stat
import tarfile
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Optional

# CNAI_ARTIFACT_TYPE is the media type for a model artifact manifest.
CNAI_ARTIFACT_TYPE = "application/vnd.cncf.model.manifest.v1+json"

# CNAI_CONFIG_MEDIA_TYPE is the media type for a model config object.
CNAI_CONFIG_MEDIA_TYPE = "application/vnd.cncf.model.config.v1+json"

# OCI_MANIFEST_MEDIA_TYPE is the standard OCI image manifest media type.
OCI_MANIFEST_MEDIA_TYPE = "application/vnd.oci.image.manifest.v1+json"

# LAYER_ANNOTATION_FILEPATH specifies the file path of the layer (string).
LAYER_ANNOTATION_FILEPATH = "org.cncf.model.filepath"

# LAYER_ANNOTATION_FILE_METADATA specifies file metadata JSON for the layer (string).
LAYER_ANNOTATION_FILE_METADATA = "org.cncf.model.file.metadata+json"

# LAYER_ANNOTATION_FILE_MEDIATYPE_UNTESTED indicates media type classification is untested (string).
LAYER_ANNOTATION_FILE_MEDIATYPE_UNTESTED = "org.cncf.model.file.mediatype.untested"

# ALLOWED_LAYER_MEDIA_TYPES contains CNAI layer media types from the model format spec.
ALLOWED_LAYER_MEDIA_TYPES = {
    "application/vnd.cncf.model.weight.v1.raw",
    "application/vnd.cncf.model.weight.v1.tar",
    "application/vnd.cncf.model.weight.v1.tar+gzip",
    "application/vnd.cncf.model.weight.v1.tar+zstd",
    "application/vnd.cncf.model.weight.config.v1.raw",
    "application/vnd.cncf.model.weight.config.v1.tar",
    "application/vnd.cncf.model.weight.config.v1.tar+gzip",
    "application/vnd.cncf.model.weight.config.v1.tar+zstd",
    "application/vnd.cncf.model.doc.v1.raw",
    "application/vnd.cncf.model.doc.v1.tar",
    "application/vnd.cncf.model.doc.v1.tar+gzip",
    "application/vnd.cncf.model.doc.v1.tar+zstd",
    "application/vnd.cncf.model.code.v1.raw",
    "application/vnd.cncf.model.code.v1.tar",
    "application/vnd.cncf.model.code.v1.tar+gzip",
    "application/vnd.cncf.model.code.v1.tar+zstd",
    "application/vnd.cncf.model.dataset.v1.raw",
    "application/vnd.cncf.model.dataset.v1.tar",
    "application/vnd.cncf.model.dataset.v1.tar+gzip",
    "application/vnd.cncf.model.dataset.v1.tar+zstd",
}

_MEDIATYPE_UNTESTED_VALUES = {"true", "false"}


def _require(condition: bool, message: str) -> None:
    if not condition:
        raise ValueError(message)


def _require_str(value: Any, message: str) -> str:
    if not isinstance(value, str) or not value:
        raise ValueError(message)
    return value


def _require_int(value: Any, message: str) -> int:
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValueError(message)
    return value


def _typeflag_for_mode(mode: int) -> int:
    if stat.S_ISDIR(mode):
        return ord(tarfile.DIRTYPE)
    if stat.S_ISLNK(mode):
        return ord(tarfile.SYMTYPE)
    return ord(tarfile.REGTYPE)


def normalize_layer_filepath(value: str) -> str:
    value = _require_str(value, "layer annotation filepath must be a non-empty string")
    if os.path.isabs(value):
        raise ValueError("layer annotation filepath must be relative")
    normalized = os.path.normpath(value).lstrip(os.sep)
    if normalized in {"", "."} or normalized.startswith(".."):
        raise ValueError("layer annotation filepath must not escape the layer")
    return normalized


@dataclass(frozen=True)
class FileMetadata:
    name: str
    mode: int
    uid: int
    gid: int
    size: int
    mtime: str
    typeflag: int

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "FileMetadata":
        if not isinstance(data, dict):
            raise ValueError("file metadata must be a JSON object")
        name = _require_str(data.get("name"), "file metadata name is required")
        mode = _require_int(data.get("mode"), "file metadata mode is required")
        uid = _require_int(data.get("uid"), "file metadata uid is required")
        gid = _require_int(data.get("gid"), "file metadata gid is required")
        size = _require_int(data.get("size"), "file metadata size is required")
        mtime = _require_str(data.get("mtime"), "file metadata mtime is required")
        typeflag = _require_int(data.get("typeflag"), "file metadata typeflag is required")
        return cls(
            name=name,
            mode=mode,
            uid=uid,
            gid=gid,
            size=size,
            mtime=mtime,
            typeflag=typeflag,
        )

    @classmethod
    def from_json(cls, value: str) -> "FileMetadata":
        try:
            data = json.loads(value)
        except json.JSONDecodeError as exc:
            raise ValueError("file metadata annotation must be valid JSON") from exc
        return cls.from_dict(data)

    @classmethod
    def from_path(cls, path: str, *, name: Optional[str] = None) -> "FileMetadata":
        stat_result = os.stat(path, follow_symlinks=False)
        mtime = datetime.fromtimestamp(stat_result.st_mtime, tz=timezone.utc).isoformat().replace("+00:00", "Z")
        return cls(
            name=name or os.path.basename(path),
            mode=stat.S_IMODE(stat_result.st_mode),
            uid=stat_result.st_uid,
            gid=stat_result.st_gid,
            size=stat_result.st_size,
            mtime=mtime,
            typeflag=_typeflag_for_mode(stat_result.st_mode),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "mode": self.mode,
            "uid": self.uid,
            "gid": self.gid,
            "size": self.size,
            "mtime": self.mtime,
            "typeflag": self.typeflag,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), separators=(",", ":"))


def is_cncf_artifact_manifest(manifest: dict[str, Any]) -> bool:
    artifact_type = manifest.get("artifactType")
    config_media = (manifest.get("config") or {}).get("mediaType", "")
    if artifact_type == CNAI_ARTIFACT_TYPE or config_media == CNAI_CONFIG_MEDIA_TYPE:
        return True
    layers = manifest.get("layers") or manifest.get("blobs") or []
    return any(layer.get("mediaType") in ALLOWED_LAYER_MEDIA_TYPES for layer in layers)


@dataclass
class Descriptor:
    media_type: str
    digest: str
    size: int
    annotations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict[str, Any], *, allowed_media_types: Optional[set[str]] = None) -> "Descriptor":
        media_type = _require_str(data.get("mediaType"), "descriptor mediaType is required")
        digest = _require_str(data.get("digest"), "descriptor digest is required")
        size = data.get("size")

        if size is None:
            raise ValueError("descriptor size is required")

        if allowed_media_types is not None:
            _require(
                media_type in allowed_media_types,
                f"descriptor mediaType '{media_type}' not in allowed set",
            )

        annotations = data.get("annotations") or {}
        if not isinstance(annotations, dict):
            raise ValueError("descriptor annotations must be a map")
        for key, value in annotations.items():
            if not isinstance(key, str) or not isinstance(value, str):
                raise ValueError("descriptor annotations must be a string map")
        return cls(media_type=media_type, digest=digest, size=int(size), annotations=annotations)

    def to_dict(self) -> dict[str, Any]:
        data: dict[str, Any] = {
            "mediaType": self.media_type,
            "digest": self.digest,
            "size": self.size,
        }
        if self.annotations:
            data["annotations"] = self.annotations
        return data

    def filepath(self) -> Optional[str]:
        value = self.annotations.get(LAYER_ANNOTATION_FILEPATH)
        if value is None:
            return None
        return normalize_layer_filepath(value)

    def file_metadata(self) -> Optional[FileMetadata]:
        value = self.annotations.get(LAYER_ANNOTATION_FILE_METADATA)
        if value is None:
            return None
        return FileMetadata.from_json(value)

    def media_type_untested(self) -> Optional[bool]:
        value = self.annotations.get(LAYER_ANNOTATION_FILE_MEDIATYPE_UNTESTED)
        if value is None:
            return None
        if value not in _MEDIATYPE_UNTESTED_VALUES:
            raise ValueError("layer annotation mediatype.untested must be 'true' or 'false'")
        return value == "true"


@dataclass
class Manifest:
    schema_version: int
    media_type: str
    artifact_type: str
    config: Descriptor
    layers: list[Descriptor]
    annotations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Manifest":
        schema_version = data.get("schemaVersion", 2)
        media_type = data.get("mediaType") or OCI_MANIFEST_MEDIA_TYPE
        artifact_type = _require_str(data.get("artifactType"), "artifactType is required")
        _require(artifact_type == CNAI_ARTIFACT_TYPE, f"artifactType must be '{CNAI_ARTIFACT_TYPE}'")
        _require(media_type == OCI_MANIFEST_MEDIA_TYPE, f"mediaType must be '{OCI_MANIFEST_MEDIA_TYPE}'")

        config_dict = data.get("config") or {}
        config = Descriptor.from_dict(config_dict)
        _require(
            config.media_type == CNAI_CONFIG_MEDIA_TYPE,
            f"config mediaType must be '{CNAI_CONFIG_MEDIA_TYPE}'",
        )

        layers_data = data.get("layers") or []
        _require(
            isinstance(layers_data, list) and len(layers_data) > 0,
            "layers must be a non-empty list",
        )
        layers = [Descriptor.from_dict(layer, allowed_media_types=ALLOWED_LAYER_MEDIA_TYPES) for layer in layers_data]

        annotations = data.get("annotations") or {}
        return cls(
            schema_version=schema_version,
            media_type=media_type,
            artifact_type=artifact_type,
            config=config,
            layers=layers,
            annotations=annotations,
        )

    def to_dict(self) -> dict[str, Any]:
        data: dict[str, Any] = {
            "schemaVersion": self.schema_version,
            "mediaType": self.media_type,
            "artifactType": self.artifact_type,
            "config": self.config.to_dict(),
            "layers": [layer.to_dict() for layer in self.layers],
        }
        if self.annotations:
            data["annotations"] = self.annotations
        return data



from __future__ import annotations

import os
from functools import lru_cache
from pathlib import Path
from typing import Literal, Optional, TypedDict, Union, cast

from ramalama.common import SemVer, engine_version
from ramalama.config import SUPPORTED_ENGINES, ActiveConfig
from ramalama.model_store.store import ModelStore
from ramalama.oci_tools import OciRef
from ramalama.transports.oci import resolver as oci_resolver
from ramalama.transports.oci.strategies import (
    BaseArtifactStrategy,
    BaseImageStrategy,
    DockerImageStrategy,
    HttpArtifactStrategy,
    PodmanArtifactStrategy,
    PodmanImageStrategy,
)

PODMAN_MIN_ARTIFACT_VERSION = SemVer(5, 7, 0)


def get_engine_image_strategy(engine: str, engine_name: SUPPORTED_ENGINES) -> type[BaseImageStrategy]:
    if engine_name == "docker":
        return DockerImageStrategy
    elif engine_name == "podman":
        return PodmanImageStrategy
    else:
        raise ValueError(f"No engine image strategies for `{engine_name}` engine.")


def get_engine_artifact_strategy(engine: str, engine_name: SUPPORTED_ENGINES) -> type[BaseArtifactStrategy]:
    if engine_name == "podman":
        version = engine_version(engine)
        if version >= PODMAN_MIN_ARTIFACT_VERSION:
            return PodmanArtifactStrategy

    return HttpArtifactStrategy


class StrategiesType(TypedDict):
    image: BaseImageStrategy
    artifact: BaseArtifactStrategy


@lru_cache
def get_strategy(
    engine: str, engine_name: SUPPORTED_ENGINES, model_store: ModelStore, kind: Literal['image', 'artifact']
) -> BaseArtifactStrategy | BaseImageStrategy:
    cls_generator = get_engine_image_strategy if kind == 'image' else get_engine_artifact_strategy
    cls = cls_generator(engine, engine_name)
    return cls(engine=engine, model_store=model_store)


class OCIStrategyFactory:
    """Resolve reference kind and return the appropriate strategy implementation."""

    def __init__(
        self,
        engine: Union[SUPPORTED_ENGINES, Path, str, None],
        model_store: ModelStore,
    ):
        if (engine := engine or ActiveConfig().engine) is None:
            raise Exception("OCIStrategyFactory require a valid engine")

        self.engine = str(engine)
        self.engine_name: SUPPORTED_ENGINES = cast(SUPPORTED_ENGINES, os.path.basename(self.engine))
        self.model_store = model_store
        self._type_resolver = oci_resolver.OCITypeResolver(self.engine, model_store=self.model_store)

    def strategies(self, kind: Literal['image', 'artifact']) -> BaseArtifactStrategy | BaseImageStrategy:
        return get_strategy(self.engine, self.engine_name, self.model_store, kind)

    def resolve_kind(self, model: OciRef) -> Optional[Literal["image", "artifact"]]:
        kind = self._type_resolver.resolve(model)
        if kind == "unknown":
            return None
        return kind

    def resolve(self, model: OciRef) -> BaseArtifactStrategy | BaseImageStrategy:
        kind = self.resolve_kind(model)
        if kind is None:
            raise ValueError(f"{model} does not exist.")

        return self.strategies(kind)



from __future__ import annotations

import json
from dataclasses import asdict
from functools import cached_property
from pathlib import Path
from typing import Union

from ramalama.benchmarks.errors import MissingStorageFolderError
from ramalama.benchmarks.schemas import BenchmarkRecord, DeviceInfoV1, get_benchmark_record
from ramalama.benchmarks.utilities import parse_jsonl

SCHEMA_VERSION = 1
BENCHMARKS_FILENAME = "benchmarks.jsonl"


class BenchmarksManager:
    def __init__(self, storage_folder: Union[str, Path, None]):
        if storage_folder is None:
            raise MissingStorageFolderError

        self.storage_folder = Path(storage_folder)
        self.storage_file = self.storage_folder / BENCHMARKS_FILENAME
        self.storage_file.parent.mkdir(parents=True, exist_ok=True)

    @cached_property
    def device_info(self) -> DeviceInfoV1:
        return DeviceInfoV1.current_device_info()

    def save(self, results: list[BenchmarkRecord] | BenchmarkRecord):
        if not isinstance(results, list):
            results = [results]

        if len(results) == 0:
            return

        with self.storage_file.open("a", encoding="utf-8") as handle:
            for record in results:
                handle.write(json.dumps(asdict(record), ensure_ascii=True))
                handle.write("\n")

    def list(self) -> list[BenchmarkRecord]:
        """List benchmark results from JSONL storage."""
        if not self.storage_file.exists():
            return []
        content = self.storage_file.read_text(encoding="utf-8")
        return [get_benchmark_record(result) for result in parse_jsonl(content)]



from __future__ import annotations

import argparse
import copy
import json
import os
import platform
import shutil
import tempfile
from dataclasses import asdict
from datetime import datetime, timezone
from http.client import HTTPConnection
from typing import Any, Optional, get_args
from urllib.parse import urlparse

from ramalama.benchmarks.manager import BenchmarksManager
from ramalama.benchmarks.schemas import BenchmarkRecord, get_benchmark_record
from ramalama.benchmarks.utilities import parse_json, print_bench_results
from ramalama.cli import (
    CoerceToBool,
    OverrideDefaultAction,
    _get_source_model,
    _rag_args,
    add_network_argument,
    default_image,
    default_rag_image,
    default_tools_image,
    get_shortnames,
    local_images,
    local_models,
    runtime_options,
    suppressCompleter,
)
from ramalama.common import (
    accel_image,
    ensure_image,
    get_gpu_type_env_vars,
    run_cmd,
    set_accel_env_vars,
    set_gpu_type_env_vars,
    version_tagged_image,
)
from ramalama.config import GGUF_QUANTIZATION_MODES, ActiveConfig, DefaultConfig
from ramalama.engine import Engine, dry_run, image_inspect
from ramalama.logger import logger
from ramalama.path_utils import file_uri_to_path
from ramalama.plugins.loader import assemble_command
from ramalama.plugins.runtimes.inference.common import ContainerizedInferenceRuntimePlugin
from ramalama.plugins.runtimes.inference.llama_cpp_commands import (
    _CACHE_REUSE_DEFAULT,
    _NGL_DEFAULT,
    _THINKING_DEFAULT,
    LlamaCppCommands,
    _default_threads,
)
from ramalama.rag import RagTransport
from ramalama.transports.api import APITransport
from ramalama.transports.base import compute_serving_port
from ramalama.transports.transport_factory import New, TransportFactory


class AddPathOrUrl(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not isinstance(values, list):
            raise ValueError("AddPathOrUrl can only be used with the settings `nargs='+'`")
        setattr(namespace, self.dest, [])
        namespace.urls = []
        for value in values:
            parsed = urlparse(value)
            if parsed.scheme in ["http", "https"]:
                namespace.urls.append(value)
            else:
                getattr(namespace, self.dest).append(file_uri_to_path(value))


def get_gpu_backend_preferences(gpu_type: str) -> list[str]:
    """Returns preferred backends for a given GPU type in order of preference.
    On Windows, vulkan is not supported on WSL2, so vendor backends are preferred."""
    is_windows = platform.system() == "Windows"

    preferences = {
        "HIP_VISIBLE_DEVICES": ["vulkan", "rocm"],  # AMD: Vulkan preferred
        "CUDA_VISIBLE_DEVICES": ["cuda"],  # NVIDIA: CUDA only
        "INTEL_VISIBLE_DEVICES": ["vulkan", "sycl", "openvino"],  # Intel: Vulkan preferred
        "ASAHI_VISIBLE_DEVICES": ["vulkan"],  # Asahi: Vulkan only
        "ASCEND_VISIBLE_DEVICES": ["cann"],  # Ascend: CANN only
        "MUSA_VISIBLE_DEVICES": ["musa"],  # MUSA: MUSA only
        "GGML_VK_VISIBLE_DEVICES": ["vulkan"],  # Vulkan: Vulkan only
    }

    if is_windows:
        preferences["HIP_VISIBLE_DEVICES"] = ["rocm", "vulkan"]
        preferences["INTEL_VISIBLE_DEVICES"] = ["sycl", "vulkan", "openvino"]

    return preferences.get(gpu_type, [])


def backend_to_gpu_env(backend: str) -> str:
    """Maps a backend name to its corresponding GPU environment variable."""
    mapping = {
        "vulkan": "GGML_VK_VISIBLE_DEVICES",
        "rocm": "HIP_VISIBLE_DEVICES",
        "cuda": "CUDA_VISIBLE_DEVICES",
        "sycl": "INTEL_VISIBLE_DEVICES",
        "openvino": "OPENVINO_VISIBLE_DEVICES",
        "cann": "ASCEND_VISIBLE_DEVICES",
        "musa": "MUSA_VISIBLE_DEVICES",
        "asahi": "ASAHI_VISIBLE_DEVICES",
    }
    return mapping.get(backend, "")


def get_available_backends() -> list[str]:
    """Returns available backends based on detected GPU, in preference order.
    Always includes 'auto' as the first option."""
    set_gpu_type_env_vars()
    gpu_type = next(iter(get_gpu_type_env_vars()), "")

    if gpu_type:
        preferences = get_gpu_backend_preferences(gpu_type)
        if preferences:
            return ["auto"] + preferences

    return ["auto", "vulkan"]


_LLAMA_CPP_IMAGES: dict[str, str] = {
    "ASAHI_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/asahi"),
    "ASCEND_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/cann"),
    "CUDA_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/cuda"),
    "GGML_VK_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/ramalama"),
    "HIP_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/rocm"),
    "INTEL_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/intel-gpu"),
    "OPENVINO_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/openvino"),
    "MUSA_VISIBLE_DEVICES": version_tagged_image("quay.io/ramalama/musa"),
}


class LlamaCppPlugin(LlamaCppCommands, ContainerizedInferenceRuntimePlugin):
    @property
    def name(self) -> str:
        return "llama.cpp"

    def _convert_to_gguf(self, outdir, source_model, args):
        """Run convert_hf_to_gguf.py inside a container to produce a GGUF file."""
        with tempfile.TemporaryDirectory(prefix="RamaLama_convert_src_") as srcdir:
            ref_file = source_model.model_store.get_ref_file(source_model.model_tag)
            for file in ref_file.files:
                blob_file_path = source_model.model_store.get_blob_file_path(file.hash)
                shutil.copyfile(blob_file_path, os.path.join(srcdir, file.name))
            engine = Engine(args)
            if engine.use_docker:
                # The uid in the container must match the euid on the host for contents of the
                # volumes to be readable.
                if hasattr(os, "geteuid"):
                    # Note: geteuid() doesn't exist on Windows, but this is only needed on Unix
                    engine.add_args(f"--user={os.geteuid()}")
            engine.add_volume(srcdir, "/model")
            engine.add_volume(outdir.name, "/output", opts="rw")
            args = copy.copy(args)
            args.model = source_model
            if not args.dryrun:
                config = ActiveConfig()
                should_pull = config.pull in ["always", "missing", "newer"]
                args.tools_image = ensure_image(args.engine, args.tools_image, should_pull=should_pull)
            engine.add_args(args.tools_image)
            engine.add_args(*self._cmd_convert(args))
            if args.dryrun:
                engine.dryrun()
            else:
                engine.run()
        return self._quantize(source_model, args, outdir.name)

    def _quantize(self, source_model, args, model_dir):
        """Run llama-quantize inside a container to quantize a GGUF model."""
        engine = Engine(args)
        if engine.use_docker:
            # The uid in the container must match the euid on the host for contents of the
            # volumes to be readable.
            if hasattr(os, "geteuid"):
                # Note: geteuid() doesn't exist on Windows, but this is only needed on Unix
                engine.add_args(f"--user={os.geteuid()}")
        engine.add_volume(model_dir, "/model", opts="rw")
        if not args.dryrun:
            config = ActiveConfig()
            should_pull = config.pull in ["always", "missing", "newer"]
            args.image = ensure_image(args.engine, args.image, should_pull=should_pull)
        engine.add_args(args.image)
        args = copy.copy(args)
        args.subcommand = "quantize"
        engine.add_args(*self._cmd_quantize(args))
        if args.dryrun:
            engine.dryrun()
        else:
            engine.run()
        return f"{source_model.model_name}-{args.gguf}.gguf"

    def service_ready_check(self, conn: HTTPConnection, args: Any, model_name: Optional[str] = None) -> bool:
        container_name = f"container {args.name}" if getattr(args, 'container', None) else 'server'
        conn.request("GET", "/health")
        health_resp = conn.getresponse()
        health_resp.read()
        if health_resp.status not in (200, 404):
            logger.debug(f"{self.name} {container_name} /health {health_resp.status}: {health_resp.reason}")
            return False

        conn.request("GET", "/models")
        models_resp = conn.getresponse()
        if models_resp.status != 200:
            logger.debug(f"{self.name} {container_name} /models status code {models_resp.status}: {models_resp.reason}")
            return False

        content = models_resp.read()
        if not content:
            logger.debug(f"{self.name} {container_name} /models returned an empty response")
            return False

        body = json.loads(content)
        if "models" not in body:
            logger.debug(f"{self.name} {container_name} /models does not include a model list in the response")
            return False

        model_names = [m["name"] for m in body["models"]]
        if not model_name:
            model_name = New(args.MODEL, args).model_alias

        if model_name not in model_names:
            logger.debug(
                f'{self.name} {container_name} /models does not include "{model_name}" in the model list: {model_names}'
            )
            return False

        logger.debug(f"{self.name} {container_name} is ready")
        return True

    def get_container_image(self, config: Any, detected_gpu_type: str) -> Optional[str]:
        backend = config.backend
        if backend == "auto":
            preferences = get_gpu_backend_preferences(detected_gpu_type)
            if preferences:
                gpu_type = backend_to_gpu_env(preferences[0])
                logger.debug(f"Auto mode selected {preferences[0]} backend for {detected_gpu_type}")
            else:
                gpu_type = detected_gpu_type
        else:
            gpu_type = backend_to_gpu_env(backend)
            if detected_gpu_type:
                preferences = get_gpu_backend_preferences(detected_gpu_type)
                if preferences and backend not in preferences:
                    gpu_name = detected_gpu_type.replace("_VISIBLE_DEVICES", "")
                    logger.warning(
                        f"Backend '{backend}' may not be compatible with detected {gpu_name} GPU. "
                        f"Recommended backends for {gpu_name}: {', '.join(preferences)}"
                    )

        override = config.images.get(gpu_type) if gpu_type else None
        return override if override else _LLAMA_CPP_IMAGES.get(gpu_type, config.default_image)

    def _container_image_is_ggml(self, args: argparse.Namespace) -> bool:
        if not args.container or args.dryrun:
            return False
        image = accel_image(ActiveConfig())
        default_image = accel_image(DefaultConfig())
        if image == default_image:
            return False
        try:
            image_entrypoint = image_inspect(args, image, format="{{ .Config.Entrypoint }}")
        except Exception as e:
            logger.debug(f"Error inspecting image {image}: {e}")
            return False  # Assume default image (non-ggml)
        # Upstream llama.cpp full image uses a wrapper script as the entrypoint
        return "tools.sh" in image_entrypoint

    # --- subcommand registration ---

    def _add_backend_arg(self, parser: "argparse.ArgumentParser") -> None:
        config = ActiveConfig()
        parser.add_argument(
            "--backend",
            dest="backend",
            default=config.backend,
            choices=get_available_backends(),
            help="GPU backend to use (auto, vulkan, rocm, cuda, sycl, openvino). See man page for details.",
        )

    def _add_ngl_arg(self, parser: "argparse.ArgumentParser") -> None:
        parser.add_argument(
            "--ngl",
            dest="ngl",
            type=int,
            default=_NGL_DEFAULT,
            help="number of layers to offload to the gpu, if available",
            completer=suppressCompleter,
        )

    def _add_threads_arg(self, parser: "argparse.ArgumentParser") -> None:
        def_threads = _default_threads()
        parser.add_argument(
            "-t",
            "--threads",
            type=int,
            default=def_threads,
            help=(
                f"number of cpu threads to use, the default is {def_threads} on this system, -1 means use this default"
            ),
            completer=suppressCompleter,
        )

    def _add_inference_args(self, parser: "argparse.ArgumentParser", command: str) -> None:
        """Add llama.cpp-specific inference args to an already-created parser."""
        super()._add_inference_args(parser, command)
        config = ActiveConfig()
        parser.add_argument(
            "--temp",
            dest="temp",
            type=float,
            default=config.temp,
            help="temperature of the response from the AI model",
            completer=suppressCompleter,
        )
        self._add_backend_arg(parser)
        parser.add_argument(
            "--cache-reuse",
            dest="cache_reuse",
            type=int,
            default=_CACHE_REUSE_DEFAULT,
            help="min chunk size to attempt reusing from the cache via KV shifting",
            completer=suppressCompleter,
        )
        self._add_ngl_arg(parser)
        if command in ["run", "serve"]:
            parser.add_argument(
                "--logfile",
                dest="logfile",
                type=str,
                help="log output to a file",
                completer=suppressCompleter,
            )
            parser.add_argument(
                "--thinking",
                default=_THINKING_DEFAULT,
                help="enable/disable thinking mode in reasoning models",
                action=CoerceToBool,
            )
        self._add_threads_arg(parser)
        if command == "serve":
            parser.add_argument("--model-draft", help="Draft model", completer=local_models)
            parser.add_argument(
                "--webui",
                dest="webui",
                choices=["on", "off"],
                default="on",
                help="enable or disable the web UI (default: on)",
            )

    @staticmethod
    def _set_openvino_env(args: argparse.Namespace) -> None:
        """Set OpenVINO env vars when Intel GPU is detected (llama.cpp-specific)."""
        if not os.environ.get("INTEL_VISIBLE_DEVICES"):
            return

        openvino_env = [
            f"GGML_OPENVINO_DEVICE={os.environ.get('GGML_OPENVINO_DEVICE', 'GPU')}",
            f"GGML_OPENVINO_STATEFUL_EXECUTION={os.environ.get('GGML_OPENVINO_STATEFUL_EXECUTION', 1)}",
        ]
        args.env = openvino_env + getattr(args, "env", [])

    def handle_subcommand(self, command: str, args: argparse.Namespace) -> list[str]:
        set_accel_env_vars()
        self._set_openvino_env(args)
        return super().handle_subcommand(command, args)

    def _do_run(self, args: argparse.Namespace, model: Any) -> None:
        if getattr(args, "rag", None):
            if isinstance(model, APITransport):
                raise ValueError("ramalama run --rag is not supported for hosted API transports.")
            self._run_rag(args, model)
            return
        super()._do_run(args, model)

    def _do_serve(self, args: argparse.Namespace, model: Any) -> None:
        if getattr(args, "rag", None):
            self._serve_rag(args, model)
            return
        super()._do_serve(args, model)

    def _add_rag_args(self, parser: "argparse.ArgumentParser") -> None:
        parser.add_argument(
            "--rag", help="RAG vector database or OCI Image to be served with the model", completer=local_models
        )
        parser.add_argument(
            "--rag-image",
            default=default_rag_image(),
            help="OCI container image to run with the specified RAG data",
            action=OverrideDefaultAction,
            completer=local_images,
        )

    def _run_rag(self, args: argparse.Namespace, model: Any) -> None:
        if not args.container:
            raise ValueError("ramalama run --rag cannot be run with the --nocontainer option.")
        args = _rag_args(args)
        model = RagTransport(model, assemble_command(args.model_args), args)
        model.ensure_model_exists(args)
        embed_serve_args, embed_proc = self._start_rag_embedding_server(args)
        try:
            model.run(args, assemble_command(args))
        finally:
            from ramalama.plugins.runtimes.inference.rag.handler import _cleanup_servers

            _cleanup_servers(args, embed_serve_args, None, embed_proc, None)

    def _serve_rag(self, args: argparse.Namespace, model: Any) -> None:
        if not args.container:
            raise ValueError("ramalama serve --rag cannot be run with the --nocontainer option.")
        args = _rag_args(args)
        model = RagTransport(model, assemble_command(args.model_args), args)
        model.ensure_model_exists(args)
        embed_serve_args, embed_proc = self._start_rag_embedding_server(args)
        try:
            model.serve(args, assemble_command(args))
        finally:
            from ramalama.plugins.runtimes.inference.rag.handler import _cleanup_servers

            _cleanup_servers(args, embed_serve_args, None, embed_proc, None)

    def _start_rag_embedding_server(self, args):
        """Start a llama.cpp embedding server for RAG inference and set embed_url on args."""
        from ramalama.plugins.runtimes.inference.rag.handler import EMBEDDING_MODEL, _build_serve_args, _wait_for_server

        embedding_model = EMBEDDING_MODEL
        set_accel_env_vars()

        embed_port = compute_serving_port(args, quiet=True, exclude=[args.port, args.model_args.port])
        embed_serve_args = _build_serve_args(args.model_args, embedding_model, embed_port, runtime_args=["--embedding"])
        # Internal models should always be pullable regardless of the user's --pull flag
        embed_serve_args.pull = ActiveConfig().pull

        embed_transport = New(embedding_model, embed_serve_args)
        if isinstance(embed_transport, APITransport):
            raise ValueError(
                f"embedding model {embedding_model} resolved to an API transport, which cannot serve locally"
            )
        embed_transport.ensure_model_exists(embed_serve_args)

        embed_cmd = assemble_command(embed_serve_args)
        embed_proc = embed_transport.serve_nonblocking(embed_serve_args, embed_cmd)

        if not args.dryrun:
            _wait_for_server(self, embed_serve_args, embed_transport.model_alias)

        if args.model_args.engine == "podman":
            embed_host = "host.containers.internal"
        else:
            embed_host = f"host.{args.model_args.engine}.internal"

        args.embed_url = f"http://{embed_host}:{embed_port}"
        return embed_serve_args, embed_proc

    def _register_run_subcommand(self, subparsers: "argparse._SubParsersAction") -> "argparse.ArgumentParser":
        parser = super()._register_run_subcommand(subparsers)
        self._add_rag_args(parser)
        return parser

    def _register_serve_subcommand(self, subparsers: "argparse._SubParsersAction") -> "argparse.ArgumentParser":
        parser = super()._register_serve_subcommand(subparsers)
        self._add_rag_args(parser)
        return parser

    def register_subcommands(self, subparsers: "argparse._SubParsersAction") -> None:
        super().register_subcommands(subparsers)

        # bench / benchmark
        bench_parser = subparsers.add_parser("bench", aliases=["benchmark"], help="benchmark specified AI Model")
        runtime_options(bench_parser, "bench")
        self._add_backend_arg(bench_parser)
        self._add_ngl_arg(bench_parser)
        self._add_threads_arg(bench_parser)
        bench_parser.add_argument(
            "--runtime-args",
            dest="runtime_args",
            default="",
            type=str,
            help="arguments to add to runtime invocation",
            completer=suppressCompleter,
        )
        bench_parser.add_argument("MODEL", completer=local_models)
        bench_parser.add_argument(
            "--format",
            choices=["table", "json"],
            default="table",
            help="output format (table or json)",
        )
        bench_parser.set_defaults(func=self._bench_handler)

        # perplexity
        perplexity_parser = subparsers.add_parser("perplexity", help="calculate perplexity for specified AI Model")
        runtime_options(perplexity_parser, "perplexity")
        self._add_inference_args(perplexity_parser, "perplexity")
        perplexity_parser.add_argument("MODEL", completer=local_models)
        perplexity_parser.set_defaults(func=self._perplexity_handler)

        # convert
        config = ActiveConfig()
        convert_parser = subparsers.add_parser(
            "convert",
            help="convert AI Model from local storage to OCI Image",
            formatter_class=argparse.RawTextHelpFormatter,
        )
        convert_parser.add_argument("--carimage", default=config.carimage, help=argparse.SUPPRESS)
        convert_parser.add_argument(
            "--gguf",
            choices=get_args(GGUF_QUANTIZATION_MODES),
            nargs="?",
            const=config.gguf_quantization_mode,
            default=None,
            help=f"GGUF quantization format. If specified without value, {config.gguf_quantization_mode} is used.",
        )
        add_network_argument(convert_parser)
        convert_parser.add_argument(
            "--tools-image",
            default=default_tools_image(),
            help="Image to use for conversion to GGUF",
            action=OverrideDefaultAction,
            completer=local_images,
        )
        convert_parser.add_argument(
            "--image",
            default=default_image(),
            help="Image to use for quantization",
            action=OverrideDefaultAction,
            completer=local_images,
        )
        convert_parser.add_argument(
            "--pull",
            dest="pull",
            type=str,
            default=config.pull,
            choices=["always", "missing", "never", "newer"],
            help="pull image policy",
        )
        convert_parser.add_argument(
            "--type",
            default=config.convert_type,
            choices=["artifact", "car", "raw"],
            help="""\
type of OCI Model Image to push.

Model "artifact" stores the AI Model as an OCI Artifact.
Model "car" includes base image with the model stored in a /models subdir.
Model "raw" contains the model and a link file model.file to it stored at /.""",
        )
        convert_parser.add_argument("SOURCE")
        convert_parser.add_argument("TARGET")
        convert_parser.set_defaults(func=self._convert_handler)

        # benchmarks (manage stored results)
        config = ActiveConfig()
        storage_folder = config.benchmarks.storage_folder
        epilog = f"Storage folder: {storage_folder}" if storage_folder else "Storage folder: not configured"
        benchmarks_parser = subparsers.add_parser(
            "benchmarks",
            help="manage and view benchmark results",
            epilog=epilog,
        )
        benchmarks_parser.set_defaults(func=lambda _: benchmarks_parser.print_help())
        benchmarks_subparsers = benchmarks_parser.add_subparsers(dest="benchmarks_command", metavar="[command]")
        benchmarks_list_parser = benchmarks_subparsers.add_parser("list", help="list benchmark results")
        benchmarks_list_parser.add_argument(
            "--limit", type=int, default=None, help="limit number of results to display"
        )
        benchmarks_list_parser.add_argument("--offset", type=int, default=0, help="offset for pagination")
        benchmarks_list_parser.add_argument(
            "--format", choices=["table", "json"], default="table", help="output format (table or json)"
        )
        benchmarks_list_parser.set_defaults(func=self._benchmarks_list_handler)

        # rag
        name_map = getattr(subparsers, "_name_parser_map", {})
        if "rag" not in name_map and ActiveConfig().container:
            self._register_rag_subcommand(subparsers)

    def _register_rag_subcommand(self, subparsers: "argparse._SubParsersAction") -> None:
        from ramalama.plugins.runtimes.inference.rag.cli import register_rag_subcommand

        register_rag_subcommand(self, subparsers)

    def _convert_handler(self, args: argparse.Namespace) -> None:
        if not args.container:
            raise ValueError("convert command cannot be run with the --nocontainer option.")

        shortnames = get_shortnames()
        tgt = shortnames.resolve(args.TARGET)
        model = TransportFactory(tgt, args).create_oci()
        source_model = _get_source_model(args)
        model.convert(source_model, args)

    def _bench_handler(self, args: argparse.Namespace) -> None:
        model = New(args.MODEL, args)
        model.ensure_model_exists(args)

        if isinstance(model, APITransport):
            raise NotImplementedError("bench is not supported for hosted API transports.")

        cmd = assemble_command(args)
        set_accel_env_vars()
        output_format = getattr(args, "format", "table")

        if args.dryrun:
            if args.container:
                model.setup_container(args)
                model.setup_mounts(args)
                model.engine.add([args.image] + cmd)
                model.engine.dryrun()
            else:
                dry_run(cmd)
            return

        if args.container:
            model.setup_container(args)
            model.setup_mounts(args)
            model.engine.add([args.image] + cmd)
            result = model.engine.run_process()
        else:
            result = run_cmd(cmd, encoding="utf-8")

        try:
            bench_results = parse_json(result.stdout)
        except (json.JSONDecodeError, ValueError):
            raise ValueError(f"Could not parse benchmark output. Expected JSON but got:\n{result.stdout}")

        base_payload: dict = {
            "created_at": datetime.now(timezone.utc).isoformat(),
            "configuration": {
                "container_image": args.image,
                "container_runtime": args.engine,
                "inference_engine": args.runtime,
                "runtime_args": cmd,
            },
        }
        results: list[BenchmarkRecord] = [
            get_benchmark_record({"result": r, **base_payload}, "v1") for r in bench_results
        ]

        if output_format == "json":
            print(result.stdout)
        else:
            print_bench_results(results)

        config = ActiveConfig()
        if not config.benchmarks.disable:
            BenchmarksManager(config.benchmarks.storage_folder).save(results)

    def _perplexity_handler(self, args: argparse.Namespace) -> None:
        model = New(args.MODEL, args)
        model.ensure_model_exists(args)

        if isinstance(model, APITransport):
            raise NotImplementedError("perplexity is not supported for hosted API transports.")

        set_accel_env_vars()
        model.execute_command(assemble_command(args), args)

    def _benchmarks_list_handler(self, args: argparse.Namespace) -> None:
        config = ActiveConfig()
        bench_manager = BenchmarksManager(config.benchmarks.storage_folder)
        results = bench_manager.list()

        if not results:
            print("No benchmark results found")
            return

        if args.format == "json":
            output = [asdict(item) for item in results]
            print(json.dumps(output, indent=2, sort_keys=True))
        else:
            print_bench_results(results)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

PathStr: TypeAlias = str
SUPPORTED_ENGINES: TypeAlias = Literal["podman", "docker"]
SUPPORTED_RUNTIMES: TypeAlias = str
COLOR_OPTIONS: TypeAlias = Literal["auto", "always", "never"]



from __future__ import annotations

import copy
import os
import random
import socket
import subprocess
import sys
import time
from abc import ABC, abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

from ramalama import chat
from ramalama.common import ContainerEntryPoint
from ramalama.compose import Compose
from ramalama.config import ActiveConfig
from ramalama.engine import Engine, dry_run, is_healthy, wait_for_healthy
from ramalama.kube import Kube
from ramalama.model_inspect.base_info import ModelInfoBase
from ramalama.model_inspect.gguf_info import GGUFModelInfo
from ramalama.model_inspect.gguf_parser import GGUFInfoParser
from ramalama.model_inspect.safetensor_info import SafetensorModelInfo
from ramalama.model_inspect.safetensor_parser import SafetensorInfoParser
from ramalama.model_store.global_store import GlobalModelStore
from ramalama.model_store.store import ModelStore
from ramalama.quadlet import Quadlet

if TYPE_CHECKING:
    from ramalama.chat import ChatOperationalArgs
    from ramalama.transports.oci.oci import OCI

from ramalama.common import (
    MNT_DIR,
    MNT_FILE_DRAFT,
    exec_cmd,
    genname,
    is_split_file_model,
    perror,
    populate_volume_from_image,
    run_cmd,
    set_accel_env_vars,
)
from ramalama.logger import logger
from ramalama.path_utils import get_container_mount_path

MODEL_TYPES = ["file", "https", "http", "oci", "huggingface", "hf", "modelscope", "ms", "ollama", "rlcr"]


file_not_found = """\
RamaLama requires the "%(cmd)s" command to be installed on the host when running with --nocontainer.
RamaLama is designed to run AI Models inside of containers, where "%(cmd)s" is already installed.
Either install a package containing the "%(cmd)s" command or run the workload inside of a container.
%(error)s"""

file_not_found_in_container = """\
RamaLama requires the "%(cmd)s" command to be installed inside of the container.
RamaLama requires the server application be installed in the container images.
Either install a package containing the "%(cmd)s" command in the container or run
with the default RamaLama
$(error)s"""


class NoGGUFModelFileFound(Exception):
    pass


class SafetensorModelNotSupported(Exception):
    pass


class NoRefFileFound(Exception):
    def __init__(self, model: str, *args):
        super().__init__(*args)

        self.model = model

    def __str__(self):
        return f"No ref file found for '{self.model}'. Please pull model."


def is_oci(transport: "Transport") -> TypeGuard["OCI"]:
    """
    Type guard to determine whether a given transport is an OCI transport.

    This assumes the transport exposes a `model_type` attribute and that
    OCI-based transports set `model_type` to the string `"oci"`.
    """
    return getattr(transport, "model_type", None) == "oci"


def trim_model_name(model):
    if model.startswith("huggingface://"):
        model = model.replace("huggingface://", "hf://", 1)

    if not model.startswith("ollama://") and not model.startswith("oci://"):
        model = model.removesuffix(":latest")

    return model


class TransportBase(ABC):
    model: str
    type: str

    def __not_implemented_error(self, param):
        return NotImplementedError(f"ramalama {param} for '{type(self).__name__}' not implemented")

    def login(self, args):
        raise self.__not_implemented_error("login")

    def logout(self, args):
        raise self.__not_implemented_error("logout")

    def pull(self, args):
        raise self.__not_implemented_error("pull")

    def push(self, source_model, args):
        raise self.__not_implemented_error("push")

    @abstractmethod
    def remove(self, args) -> bool:
        raise self.__not_implemented_error("rm")

    @abstractmethod
    def exists(self) -> bool:
        raise self.__not_implemented_error("exists")

    @abstractmethod
    def inspect(self, args):
        raise self.__not_implemented_error("inspect")


class Transport(TransportBase):
    """Transport super class"""

    type: str = "Transport"

    def __init__(self, model: str, model_store_path: str):
        self.model = model

        split: list[str] = self.model.rsplit("/", 1)
        self.directory: str = split[0] if len(split) > 1 else ""
        self.filename: str = split[1] if len(split) > 1 else split[0]

        self._model_name: str
        self._model_tag: str
        self._model_organization: str
        self._model_type: str
        self._model_name, self._model_tag, self._model_organization = self.extract_model_identifiers()
        self._model_type = type(self).__name__.lower()

        self._model_store_path: str = model_store_path
        self._model_store: Optional["ModelStore"] = None

        self.draft_model: Optional[Transport] = None

    def extract_model_identifiers(self):
        model_name = self.model
        model_tag = "latest"
        model_organization = ""

        # extract model tag from name if exists
        if ":" in model_name:
            model_name, model_tag = model_name.rsplit(":", 1)

        # extract model organization from name if exists and update name
        split = model_name.rsplit("/", 1)
        model_organization = split[0].removeprefix("/") if len(split) > 1 else ""
        model_name = split[1] if len(split) > 1 else split[0]

        return model_name, model_tag, model_organization

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def model_alias(self):
        return f"{self.model_organization}/{self.model_name}"

    @property
    def model_tag(self) -> str:
        return self._model_tag

    @property
    def model_organization(self) -> str:
        return self._model_organization

    @property
    def model_type(self) -> str:
        return self._model_type

    @property
    def model_store(self) -> "ModelStore":
        if self._model_store is None:
            name, _, orga = self.extract_model_identifiers()
            self._model_store = ModelStore(GlobalModelStore(self._model_store_path), name, self.model_type, orga)
        return self._model_store

    def _get_all_model_part_paths(
        self, use_container: bool, should_generate: bool, dry_run: bool
    ) -> list[tuple[str, str]]:
        """
        Returns a list of (src_path, dest_path) tuples for all parts of a model.
        For single-file models, returns a list with one tuple.
        For multi-part models, returns a tuple for each part.
        """
        if dry_run:
            return [("/path/to/model", f"{MNT_DIR}/model.file")]

        if self.model_type == 'oci':
            # OCI models don't use this path for multi-part handling
            entry_path_src = self._get_entry_model_path(False, False, False)
            entry_path_dest = self._get_entry_model_path(True, True, False)
            return [(entry_path_src, entry_path_dest)]

        ref_file = self.model_store.get_ref_file(self.model_tag)
        if ref_file is None:
            raise NoRefFileFound(self.model)

        gguf_files = ref_file.model_files
        safetensor_files = ref_file.safetensor_model_files
        if safetensor_files:
            # Safetensor models use directory mounts, not individual files
            src_path = self.model_store.get_snapshot_directory_from_tag(self.model_tag)
            if use_container or should_generate:
                dest_path = MNT_DIR
            else:
                dest_path = src_path
            return [(src_path, dest_path)]
        elif not gguf_files:
            raise NoGGUFModelFileFound()

        model_parts = []
        for model_file in gguf_files:
            if use_container or should_generate:
                dest_path = f"{MNT_DIR}/{model_file.name}"
            else:
                dest_path = self.model_store.get_blob_file_path(model_file.hash)
            src_path = self.model_store.get_blob_file_path(model_file.hash)
            model_parts.append((src_path, dest_path))

        # Sort multi-part models by filename to ensure correct order
        if len(model_parts) > 1 and any("-00001-of-" in name for _, name in model_parts):
            model_parts.sort(key=lambda x: x[1])

        return model_parts

    def _get_entry_model_path(self, use_container: bool, should_generate: bool, dry_run: bool) -> str:
        """
        Returns the path to the model blob on the host if use_container and should_generate are both False.
        Or returns the path to the mounted file inside a container.
        """
        if dry_run:
            return "/path/to/model"

        if is_oci(self):
            if use_container or should_generate:
                return self.entrypoint_path()
            else:
                return f"oci://{self.model}"

        ref_file = self.model_store.get_ref_file(self.model_tag)
        if ref_file is None:
            raise NoRefFileFound(self.model)

        gguf_files = ref_file.model_files
        safetensor_files = ref_file.safetensor_model_files
        if safetensor_files:
            if use_container or should_generate:
                return MNT_DIR
            return self.model_store.get_snapshot_directory_from_tag(self.model_tag)
        elif not gguf_files:
            raise NoGGUFModelFileFound()

        # Use the first model file
        model_file = gguf_files[0]
        if is_split_file_model(self.model_name):
            # Find model file with index 1 for split models
            index_models = [file for file in gguf_files if "-00001-of-" in file.name]
            if len(index_models) != 1:
                raise Exception(f"Found multiple index 1 gguf models: {index_models}")
            model_file = index_models[0]

        if use_container or should_generate:
            return f"{MNT_DIR}/{model_file.name}"
        return self.model_store.get_blob_file_path(model_file.hash)

    def _get_inspect_model_path(self, dry_run: bool) -> str:
        """Return a concrete file path for inspection.
        Prefer the safetensor blob if available; otherwise use the entry path.
        """
        if dry_run:
            return "/path/to/model"
        if self.model_type == 'oci':
            return self._get_entry_model_path(False, False, dry_run)
        safetensor_blob = self.model_store.get_safetensor_blob_path(self.model_tag, self.filename)
        return safetensor_blob or self._get_entry_model_path(False, False, dry_run)

    def _get_mmproj_path(self, use_container: bool, should_generate: bool, dry_run: bool) -> Optional[str]:
        """
        Returns the path to the mmproj blob on the host if use_container and should_generate are both False.
        Or returns the path to the mounted file inside a container.
        """
        if dry_run:
            return ""

        if self.model_type == 'oci':
            return None

        ref_file = self.model_store.get_ref_file(self.model_tag)
        if ref_file is None:
            raise NoRefFileFound(self.model)

        if not ref_file.mmproj_files:
            return None

        # Use the first mmproj file
        mmproj_file = ref_file.mmproj_files[0]
        if use_container or should_generate:
            return f"{MNT_DIR}/{mmproj_file.name}"
        return self.model_store.get_blob_file_path(mmproj_file.hash)

    def _get_chat_template_path(self, use_container: bool, should_generate: bool, dry_run: bool) -> Optional[str]:
        """
        Returns the path to the chat template blob on the host if use_container and should_generate are both False.
        Or returns the path to the mounted file inside a container.
        """
        if dry_run:
            return ""

        if self.model_type == 'oci':
            return None

        ref_file = self.model_store.get_ref_file(self.model_tag)
        if ref_file is None:
            raise NoRefFileFound(self.model)

        if not ref_file.chat_templates:
            return None

        # Use the last chat template file (may have been go template converted to jinja)
        chat_template_file = ref_file.chat_templates[-1]
        if use_container or should_generate:
            return f"{MNT_DIR}/{chat_template_file.name}"
        return self.model_store.get_blob_file_path(chat_template_file.hash)

    def remove(self, args) -> bool:
        _, tag, _ = self.extract_model_identifiers()
        if self.model_store.remove_snapshot(tag):
            return True
        if not args.ignore:
            raise KeyError(f"Model '{self.model}' not found")
        return False

    def get_container_name(self, args):
        if getattr(args, "name", None):
            return args.name

        return genname()

    def new_engine(self, args) -> Engine:
        return Engine(args)

    def base(self, args, name):
        if self.type == "Ollama":
            args.UNRESOLVED_MODEL = args.MODEL
            resolve_model = getattr(self, "resolve_model", None)
            if not callable(resolve_model):
                raise NotImplementedError("Ollama transport requires resolve_model; it is missing or not callable")
            args.MODEL = resolve_model()
        self.engine = self.new_engine(args)
        if args.subcommand == "run" and not getattr(args, "ARGS", None) and sys.stdin.isatty():
            self.engine.add(["-i"])

        self.engine.add(
            [
                "--label",
                "ai.ramalama",
                "--name",
                name,
                "--env=HOME=/tmp",
                "--init",
            ]
        )

    def setup_container(self, args):
        name = self.get_container_name(args)
        self.base(args, name)

    def exec_model_in_container(self, cmd_args, args):
        if not args.container:
            return False

        if len(cmd_args) > 0 and isinstance(cmd_args[0], ContainerEntryPoint):
            # Ignore entrypoint
            cmd_args = cmd_args[1:]

        self.setup_container(args)
        self.setup_mounts(args)

        # Make sure Image precedes cmd_args
        self.engine.add([args.image] + cmd_args)

        if args.dryrun:
            self.engine.dryrun()
            return True
        # Detached serve: use run_cmd so the process returns and the plugin can run the healthcheck
        if getattr(args, "detach", False) and getattr(args, "subcommand", "") == "serve":
            run_cmd(self.engine.exec_args, ignore_all=args.noout)
            return True
        self.engine.exec(stdout2null=args.noout)
        return True

    def setup_mounts(self, args):
        if args.dryrun:
            return

        if self.model_type == 'oci':
            strategy = getattr(self, "strategy", None)
            mount_cmd_fn = getattr(self, "mount_cmd", None)
            if strategy is None or mount_cmd_fn is None:
                raise NotImplementedError("OCI transport requires strategy and mount_cmd")
            if self.engine.use_podman or strategy.kind == "artifact":
                mount_cmd = mount_cmd_fn()
            elif self.engine.use_docker:
                output_filename = self._get_entry_model_path(args.container, True, args.dryrun)
                volume = populate_volume_from_image(self, args, os.path.basename(output_filename))
                mount_cmd = mount_cmd_fn(volume, MNT_DIR)
            else:
                raise NotImplementedError(f"No compatible oci mount method for engine: {self.engine.args.engine}")
            self.engine.add([mount_cmd])
            return None

        ref_file = self.model_store.get_ref_file(self.model_tag)

        if ref_file is None:
            raise NoRefFileFound(self.model)

        # mount all files into container with file name instead of hash
        for file in ref_file.files:
            blob_path = self.model_store.get_blob_file_path(file.hash)
            # Convert path to container-friendly format (handles Windows path conversion)
            container_blob_path = get_container_mount_path(blob_path)
            mount_path = f"{MNT_DIR}/{file.name}"
            self.engine.add(
                [f"--mount=type=bind,src={container_blob_path},destination={mount_path},ro{self.engine.relabel()}"]
            )

        if self.draft_model:
            draft_model = self.draft_model._get_entry_model_path(args.container, args.generate, args.dryrun)
            # Convert path to container-friendly format (handles Windows path conversion)
            container_draft_model = get_container_mount_path(draft_model)
            mount_opts = f"--mount=type=bind,src={container_draft_model},destination={MNT_FILE_DRAFT}"
            mount_opts += f",ro{self.engine.relabel()}"
            self.engine.add([mount_opts])

    def serve_nonblocking(self, args, cmd: list[str]) -> Optional[subprocess.Popen]:
        if args.container:
            args.name = self.get_container_name(args)

        args.host = ActiveConfig().host
        args.detach = True

        set_accel_env_vars()

        if args.container:
            # For container mode, set up the container and start it with subprocess
            self.setup_container(args)
            self.setup_mounts(args)
            # Make sure Image precedes cmd_args
            self.engine.add([args.image] + cmd)

            if args.dryrun:
                self.engine.dryrun()
                return None

            # Start the container using subprocess.Popen
            process = subprocess.Popen(
                self.engine.exec_args,
            )
            return process

        # Non-container mode: run the command directly with subprocess
        if args.dryrun:
            dry_run(cmd)
            return None

        process = subprocess.Popen(
            cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return process

    def _connect_and_chat(self, args, server_process):
        """Connect to the server and start chat in the parent process."""

        args.url = f"http://127.0.0.1:{args.port}/v1"

        # Model name in the chat request must match RamalamaModelContext.alias()
        chat_args = copy.deepcopy(args)
        chat_args.model = f"{self.model_organization}/{self.model_name}"

        if args.container:
            return self._handle_container_chat(chat_args, server_process)
        else:
            # Store the Popen object for monitoring
            chat_args.server_process = server_process
            chat.chat(chat_args)
            return 0

    def chat_operational_args(self, args) -> "Optional[ChatOperationalArgs]":
        return None

    def wait_for_healthy(self, args):
        wait_for_healthy(args, partial(is_healthy, model_name=self.model_alias))

    def _handle_container_chat(self, args, server_process):
        """Handle chat for container-based execution."""

        # Wait for the server process to complete (blocking)
        exit_code = server_process.wait()
        if exit_code != 0:
            raise ValueError(f"Failed to serve model {self.model_name}, for ramalama run command")

        if not args.dryrun:
            try:
                self.wait_for_healthy(args)
            except subprocess.TimeoutExpired as e:
                logger.error(f"Failed to serve model {self.model_name}, for ramalama run command")
                logger.error(f"{e}: logs: {e.output}")
                raise

        args.ignore = getattr(args, "dryrun", False)
        for i in range(6):
            try:
                chat.chat(args, self.chat_operational_args(args))
                break
            except Exception as e:
                if i >= 5:
                    raise e
                time.sleep(1)
        return 0

    def _is_server_ready(self, port):
        """Check if the server is ready to accept connections."""
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.settimeout(2)
                result = s.connect_ex(('127.0.0.1', int(port)))
                return result == 0
        except (socket.error, ValueError):
            return False

    def _cleanup_server_process(self, process):
        """Clean up the server process."""
        if not process:
            return

        process.terminate()
        try:
            process.wait(timeout=5)
        except subprocess.TimeoutExpired:
            process.kill()

    def exists(self) -> bool:
        _, _, all = self.model_store.get_cached_files(self.model_tag)
        return all

    def ensure_model_exists(self, args):
        self.validate_args(args)

        if args.dryrun or self.exists():
            return

        if args.pull == "never":
            raise ValueError(f"{args.MODEL} does not exist")

        self.pull(args)

    def validate_args(self, args):
        # If --nocontainer=False was specified return valid
        if args.container:
            return
        if getattr(args, "privileged", None):
            raise KeyError(
                "--nocontainer and --privileged options conflict. The --privileged option requires a container."
            )
        # If --name was not specified return valid
        if not getattr(args, "name", None):
            return
        # If --generate was specified return valid
        if getattr(args, "generate", False):
            # Do not fail on serve if user specified --generate
            return

        raise KeyError("--nocontainer and --name options conflict. The --name option requires a container.")

    def generate_container_config(self, args, exec_args):
        # Get the blob paths (src) and mounted paths (dest)
        model_src_path = self._get_entry_model_path(False, False, args.dryrun)
        chat_template_src_path = self._get_chat_template_path(False, False, args.dryrun)
        mmproj_src_path = self._get_mmproj_path(False, False, args.dryrun)
        model_dest_path = self._get_entry_model_path(True, True, args.dryrun)
        chat_template_dest_path = self._get_chat_template_path(True, True, args.dryrun)
        mmproj_dest_path = self._get_mmproj_path(True, True, args.dryrun)

        # Get all model parts (for multi-part models)
        model_parts = self._get_all_model_part_paths(False, True, args.dryrun)

        if args.generate.gen_type == "quadlet":
            self.quadlet(
                (model_src_path, model_dest_path),
                (chat_template_src_path, chat_template_dest_path),
                (mmproj_src_path, mmproj_dest_path),
                args,
                exec_args,
                args.generate.output_dir,
                model_parts,
            )
        elif args.generate.gen_type == "kube":
            self.kube(
                (model_src_path.removeprefix("oci://"), model_dest_path),
                (chat_template_src_path, chat_template_dest_path),
                (mmproj_src_path, mmproj_dest_path),
                args,
                exec_args,
                args.generate.output_dir,
            )
        elif args.generate.gen_type == "quadlet/kube":
            self.quadlet_kube(
                (model_src_path, model_dest_path),
                (chat_template_src_path, chat_template_dest_path),
                (mmproj_src_path, mmproj_dest_path),
                args,
                exec_args,
                args.generate.output_dir,
                model_parts,
            )
        elif args.generate.gen_type == "compose":
            self.compose(
                (model_src_path, model_dest_path),
                (chat_template_src_path, chat_template_dest_path),
                (mmproj_src_path, mmproj_dest_path),
                args,
                exec_args,
                args.generate.output_dir,
            )

    def execute_command(self, exec_args, args):
        try:
            if self.exec_model_in_container(exec_args, args):
                return
            if args.dryrun:
                dry_run(exec_args)
                return
            exec_cmd(exec_args, stdout2null=args.noout, stderr2null=args.noout)
        except FileNotFoundError as e:
            if args.container:
                raise NotImplementedError(
                    file_not_found_in_container % {"cmd": exec_args[0], "error": str(e).strip("'")}
                )
            raise NotImplementedError(file_not_found % {"cmd": exec_args[0], "error": str(e).strip("'")})

    def quadlet(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir, model_parts=None):
        quadlet = Quadlet(
            self.model_name,
            model_paths,
            chat_template_paths,
            mmproj_paths,
            args,
            exec_args,
            self.is_artifact,
            model_parts,
        )
        for generated_file in quadlet.generate():
            generated_file.write(output_dir)

    def quadlet_kube(
        self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir, model_parts=None
    ):
        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.is_artifact)
        kube.generate().write(output_dir)

        quadlet = Quadlet(
            kube.name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.is_artifact, model_parts
        )
        quadlet.kube().write(output_dir)

    def kube(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
        kube = Kube(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args, self.is_artifact)
        kube.generate().write(output_dir)

    def compose(self, model_paths, chat_template_paths, mmproj_paths, args, exec_args, output_dir):
        compose = Compose(self.model_name, model_paths, chat_template_paths, mmproj_paths, args, exec_args)
        compose.generate().write(output_dir)

    def inspect_metadata(self) -> dict[str, Any]:
        model_path = self._get_entry_model_path(False, False, False)

        if GGUFInfoParser.is_model_gguf(model_path):
            return GGUFInfoParser.parse_metadata(model_path).data
        return {}

    def inspect(
        self,
        show_all: bool = False,
        show_all_metadata: bool = False,
        get_field: str = "",
        as_json: bool = False,
        dryrun: bool = False,
    ) -> Any:
        model_name = self.filename
        model_registry = self.type.lower()
        model_path = self._get_inspect_model_path(dryrun)
        if GGUFInfoParser.is_model_gguf(model_path):
            if not show_all_metadata and get_field == "":
                gguf_info: GGUFModelInfo = GGUFInfoParser.parse(model_name, model_registry, model_path)
                return gguf_info.serialize(json=as_json, all=show_all)

            metadata = GGUFInfoParser.parse_metadata(model_path)
            if show_all_metadata:
                return metadata.serialize(json=as_json)
            elif get_field != "":  # If a specific field is requested, print only that field
                field_value = metadata.get(get_field)
                if field_value is None:
                    raise KeyError(f"Field '{get_field}' not found in GGUF model metadata")
                return field_value

        if SafetensorInfoParser.is_model_safetensor(model_name):
            safetensor_info: SafetensorModelInfo = SafetensorInfoParser.parse(model_name, model_registry, model_path)
            return safetensor_info.serialize(json=as_json, all=show_all)

        return ModelInfoBase(model_name, model_registry, model_path).serialize(json=as_json)

    def print_pull_message(self, model_name) -> None:
        model_name = trim_model_name(model_name)
        # Write messages to stderr
        perror(f"Downloading {model_name} ...")
        perror(f"Trying to pull {model_name} ...")

    @property
    def is_artifact(self) -> bool:
        return False


def compute_ports(exclude: Optional[list[str]] = None) -> list[int]:
    excluded = set() if exclude is None else set(map(int, exclude))

    port_range = ActiveConfig().default_port_range
    ports = [p for p in range(port_range[0], port_range[1] + 1) if p not in excluded]

    if not ports:
        raise ValueError("All ports in the default port range were exhausted by the exclusion list.")

    first_port = ports.pop(0)
    random.shuffle(ports)
    # try always the first port before the randomized others
    return [first_port] + ports


def get_available_port_if_any(exclude: Optional[list[str]] = None) -> int:
    ports = compute_ports(exclude=exclude)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        chosen_port = 0
        for target_port in ports:
            logger.debug(f"Checking if {target_port} is available")
            try:
                s.bind(('localhost', target_port))
            except OSError:
                continue
            else:
                chosen_port = target_port
                break
        return chosen_port


def compute_serving_port(args, quiet: bool = False, exclude: Optional[list[str]] = None) -> str:
    # user probably specified a custom port, don't override the choice
    if hasattr(args, 'port_override'):
        target_port = args.port
    else:
        # otherwise compute a random serving port in the range
        target_port = get_available_port_if_any(exclude=exclude)

    if target_port == 0:
        raise IOError("no available port could be detected. Please ensure you have enough free ports.")
    if not quiet:
        openai = f"http://localhost:{target_port}"
        if getattr(args, "api", None) == "llama-stack":
            perror(f"Llama Stack RESTAPI: {openai}")
            openai = openai + "/v1/openai"
            perror(f"OpenAI RESTAPI: {openai}")
    return str(target_port)



from __future__ import annotations

import glob
import hashlib
import json
import os
import platform
import random
import re
import shutil
import string
import subprocess
import sys
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Literal, Optional, Protocol, TypedDict, Union, cast, get_args

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
import yaml

import ramalama.amdkfd as amdkfd
from ramalama.logger import logger
from ramalama.version import version

if TYPE_CHECKING:
    from argparse import Namespace

    from ramalama.arg_types import SUPPORTED_ENGINES, ContainerArgType
    from ramalama.config import Config
    from ramalama.transports.base import Transport

MNT_DIR = "/mnt/models"
MNT_FILE = f"{MNT_DIR}/model.file"
MNT_MMPROJ_FILE = f"{MNT_DIR}/mmproj.file"
MNT_FILE_DRAFT = f"{MNT_DIR}/draft_model.file"
MNT_CHAT_TEMPLATE_FILE = f"{MNT_DIR}/chat_template.file"

RAG_DIR = "/rag"
RAG_CONTENT = f"{MNT_DIR}/vector.db"

MIN_VRAM_BYTES = 1073741824  # 1GiB

SPLIT_MODEL_PATH_RE = r'(.*?)(?:/)?([^/]*)-00001-of-(\d{5})\.gguf'


def is_split_file_model(model_path):
    """returns true if ends with -%05d-of-%05d.gguf"""
    return bool(re.match(SPLIT_MODEL_PATH_RE, model_path))


def sanitize_filename(filename: str) -> str:
    return filename.replace(":", "-")


podman_machine_accel = False


def confirm_no_gpu(name, provider) -> bool:
    while True:
        user_input = (
            input(
                f"Warning! Your VM {name} is using {provider}, which does not support GPU. "
                "Only the provider libkrun has GPU support. "
                "See `man ramalama-macos` for more information. "
                "Do you want to proceed without GPU? (yes/no): "
            )
            .strip()
            .lower()
        )
        if user_input in ["yes", "y"]:
            return True
        if user_input in ["no", "n"]:
            return False
        print("Invalid input. Please enter 'yes' or 'no'.")


def handle_provider(machine, config: Optional[Config] = None) -> Optional[bool]:
    global podman_machine_accel
    name = machine.get("Name")
    provider = machine.get("VMType")
    running = machine.get("Running")
    if running:
        if provider == "applehv":
            if config is not None and config.user.no_missing_gpu_prompt:
                return True
            else:
                return confirm_no_gpu(name, provider)
        if "krun" in provider:
            podman_machine_accel = True
            return True

    return None


def apple_vm(engine: SUPPORTED_ENGINES, config: Optional[Config] = None) -> bool:
    podman_machine_list = [engine, "machine", "list", "--format", "json", "--all-providers"]
    try:
        machines_json = run_cmd(podman_machine_list, ignore_stderr=True, encoding="utf-8").stdout.strip()
        machines = json.loads(machines_json)
        for machine in machines:
            result = handle_provider(machine, config)
            if result is not None:
                return result
    except (subprocess.CalledProcessError, json.JSONDecodeError, FileNotFoundError) as e:
        logger.warning(f"Failed to list and parse podman machines: {e}")
    return False


def perror(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def available(cmd: str) -> bool:
    return shutil.which(cmd) is not None


def quoted(arr) -> str:
    """Return string with quotes around elements containing spaces."""
    return " ".join(['"' + element + '"' if ' ' in element else element for element in arr])


def exec_cmd(args, stdout2null: bool = False, stderr2null: bool = False):
    logger.debug(f"exec_cmd: {quoted(args)}")

    stdout_target = subprocess.DEVNULL if stdout2null else None
    stderr_target = subprocess.DEVNULL if stderr2null else None
    try:
        result = subprocess.run(args, stdout=stdout_target, stderr=stderr_target, check=False)
        sys.exit(result.returncode)
    except Exception as e:
        perror(f"Failed to execute {quoted(args)}: {e}")
        raise


def run_cmd(
    args: Sequence[str],
    cwd: Optional[str] = None,
    stdout: Union[int, IO[Any], None] = subprocess.PIPE,
    stdin: Union[int, IO[Any], None] = subprocess.DEVNULL,
    ignore_stderr: bool = False,
    ignore_all: bool = False,
    encoding: Optional[str] = None,
    env: Optional[dict[str, str]] = None,
) -> subprocess.CompletedProcess[Any]:
    """
    Run the given command arguments.

    Args:
    args: command line arguments to execute in a subprocess
    cwd: optional working directory to run the command from
    stdout: standard output configuration
    ignore_stderr: if True, ignore standard error
    ignore_all: if True, ignore both standard output and standard error
    encoding: encoding to apply to the result text
    """
    logger.debug(f"run_cmd: {quoted(args)}")
    logger.debug(f"Working directory: {cwd}")
    logger.debug(f"Ignore stderr: {ignore_stderr}")
    logger.debug(f"Ignore all: {ignore_all}")
    logger.debug(f"env: {env}")

    serr = None
    if ignore_all or ignore_stderr:
        serr = subprocess.DEVNULL

    sout = stdout
    if ignore_all:
        sout = subprocess.DEVNULL

    if env:
        env = os.environ | env

    result = subprocess.run(
        args, check=True, cwd=cwd, stdout=sout, stderr=serr, stdin=stdin, encoding=encoding, env=env
    )
    logger.debug(f"Command finished with return code: {result.returncode}")

    return result


def populate_volume_from_image(model: Transport, args: Namespace, output_filename: str, src_model_dir: str = "models"):
    """Builds a Docker-compatible mount string that mirrors Podman image mounts for model assets.

    This function requires the model
    """

    vol_hash = hashlib.sha256(model.model.encode()).hexdigest()[:12]
    volume = f"ramalama-models-{vol_hash}"
    src = f"src-{vol_hash}"

    # Ensure volume exists
    run_cmd([args.engine, "volume", "create", volume], ignore_stderr=True)

    # Fresh source container to export from
    run_cmd([args.engine, "rm", "-f", src], ignore_stderr=True)
    run_cmd([args.engine, "create", "--name", src, model.model])

    try:
        # Stream whole rootfs -> extract only models/<basename>
        export_cmd = [args.engine, "export", src]
        untar_cmd = [
            args.engine,
            "run",
            "--rm",
            "-i",
            "--mount",
            f"type=volume,src={volume},dst=/mnt",
            "busybox",
            "tar",
            "-C",
            "/mnt",
            "--strip-components=1",
            "-x",
            "-p",
            "-f",
            "-",
            f"{src_model_dir}/{output_filename}",  # NOTE: double check this
        ]

        with (
            subprocess.Popen(export_cmd, stdout=subprocess.PIPE) as p_out,
            subprocess.Popen(untar_cmd, stdin=p_out.stdout) as p_in,
        ):
            p_out.stdout.close()  # type: ignore
            rc_in = p_in.wait()
            rc_out = p_out.wait()
            if rc_in != 0 or rc_out != 0:
                raise subprocess.CalledProcessError(rc_in or rc_out, untar_cmd if rc_in else export_cmd)
    finally:
        run_cmd([args.engine, "rm", "-f", src], ignore_stderr=True)

    return volume


def generate_sha256_binary(to_hash: bytes, with_sha_prefix: bool = True) -> str:
    """
    Generates a sha256 for data bytes.

    Args:
    to_hash (bytes): The data to generate the sha256 hash for.

    Returns:
    str: Hex digest of the input appended to the prefix sha256-
    """
    h = hashlib.new("sha256")
    h.update(to_hash)
    if with_sha_prefix:
        return f"sha256-{h.hexdigest()}"
    return h.hexdigest()


def generate_sha256(to_hash: str, with_sha_prefix: bool = True) -> str:
    return generate_sha256_binary(to_hash.encode("utf-8"), with_sha_prefix)


def verify_checksum(filename: str) -> bool:
    """
    Verifies if the SHA-256 checksum of a file matches the checksum provided in
    the filename.

    Args:
    filename (str): The filename containing the checksum prefix
                    (e.g., "sha256:<checksum>")

    Returns:
    bool: True if the checksum matches, False otherwise.
    """

    if not os.path.exists(filename):
        return False

    # Check if the filename starts with "sha256:" or "sha256-" and extract the checksum from filename
    expected_checksum = ""
    fn_base = os.path.basename(filename)
    if fn_base.startswith("sha256:"):
        expected_checksum = fn_base.split(":")[1]
    elif fn_base.startswith("sha256-"):
        expected_checksum = fn_base.split("-")[1]
    else:
        raise ValueError(f"filename has to start with 'sha256:' or 'sha256-': {fn_base}")

    if len(expected_checksum) != 64:
        raise ValueError("invalid checksum length in filename")

    # Calculate the SHA-256 checksum of the file contents
    sha256_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    # Compare the checksums
    return sha256_hash.hexdigest() == expected_checksum


def genname():
    return "ramalama-" + "".join(random.choices(string.ascii_letters + string.digits, k=10))


@lru_cache
def engine_version(engine: SUPPORTED_ENGINES | Path | str) -> SemVer:
    # Create manifest list for target with imageid
    cmd_args = [str(engine), "version", "--format", "{{ .Client.Version }}"]
    return SemVer.parse(run_cmd(cmd_args, encoding="utf-8").stdout.strip())


class CDI_DEVICE(TypedDict):
    name: str


class CDI_RETURN_TYPE(TypedDict):
    devices: list[CDI_DEVICE]


def load_cdi_config(spec_dirs: list[str]) -> Optional[CDI_RETURN_TYPE]:
    # Load and merge all CDI configuration files from the given directories.
    # When multiple CDI configs exist (e.g. /var/run/cdi and /etc/cdi), all are
    # merged so that device "all" and other devices are found regardless of
    # which file they appear in (fixes #2485).
    merged_devices: list[CDI_DEVICE] = []
    seen_names: set[str] = set()

    for spec_dir in spec_dirs:
        if not os.path.isdir(spec_dir):
            continue
        for root, _, files in os.walk(spec_dir):
            for file in sorted(files):
                _, ext = os.path.splitext(file)
                if ext not in (".yaml", ".yml", ".json"):
                    continue
                file_path = os.path.join(root, file)
                config = None
                try:
                    with open(file_path, "r") as stream:
                        if ext == ".json":
                            config = json.load(stream)
                        else:
                            config = yaml.safe_load(stream)
                except (OSError, yaml.YAMLError, json.JSONDecodeError, UnicodeDecodeError) as e:
                    logger.warning(f"Failed to load/parse CDI file {file_path}: {e}")
                    continue
                if not config or not isinstance(config, dict):
                    continue
                for cdi_device in config.get("devices", []):
                    if isinstance(cdi_device, dict) and (name := cdi_device.get("name")):
                        if name not in seen_names:
                            seen_names.add(name)
                            merged_devices.append(cast(CDI_DEVICE, cdi_device))

    if not merged_devices:
        return None
    return {"devices": merged_devices}


def get_podman_machine_cdi_config() -> Optional[CDI_RETURN_TYPE]:
    cdi_config = run_cmd(["podman", "machine", "ssh", "cat", "/etc/cdi/nvidia.yaml"], encoding="utf-8").stdout.strip()
    if cdi_config:
        return yaml.safe_load(cdi_config)
    return None


def find_in_cdi(devices: list[str]) -> tuple[list[str], list[str]]:
    # Attempts to find a CDI configuration for each device in devices
    # and returns a list of configured devices and a list of
    # unconfigured devices.
    if platform.system() == "Windows":
        cdi = get_podman_machine_cdi_config()
    else:
        cdi = load_cdi_config(['/var/run/cdi', '/etc/cdi'])
    try:
        cdi_devices = cdi.get("devices", []) if cdi else []
        cdi_device_names = [name for cdi_device in cdi_devices if (name := cdi_device.get("name"))]
    except (AttributeError, KeyError, TypeError) as e:
        # Malformed YAML or JSON. Treat everything as unconfigured but warn.
        logger.warning(f"Unable to process CDI configuration: {e}")
        return ([], devices)

    configured = []
    unconfigured = []
    for device in devices:
        if device in cdi_device_names:
            configured.append(device)
        # A device can be specified by a prefix of the uuid
        elif device.startswith("GPU") and any(name.startswith(device) for name in cdi_device_names):
            configured.append(device)
        else:
            perror(f"Device {device} does not have a CDI configuration")
            unconfigured.append(device)

    return configured, unconfigured


def check_asahi() -> Optional[Literal["asahi"]]:
    if os.path.exists('/proc/device-tree/compatible'):
        try:
            with open('/proc/device-tree/compatible', 'rb') as f:
                content = f.read().split(b"\0")
                if b"apple,arm-platform" in content:
                    os.environ["ASAHI_VISIBLE_DEVICES"] = "1"
                    return "asahi"
        except OSError:
            pass

    return None


def check_metal(args: ContainerArgType) -> bool:
    if args.container:
        return False
    return platform.system() == "Darwin"


@lru_cache(maxsize=1)
def check_nvidia() -> Optional[Literal["cuda"]]:
    try:
        command = ['nvidia-smi', '--query-gpu=index,uuid', '--format=csv,noheader']
        result = run_cmd(command, encoding="utf-8")
    except (OSError, subprocess.CalledProcessError):
        return None

    smi_lines = result.stdout.splitlines()
    parsed_lines: list[list[str]] = [[item.strip() for item in line.split(',')] for line in smi_lines if line]

    if not parsed_lines:
        return None

    indices, uuids = map(list, zip(*parsed_lines))
    # Get the list of devices specified by CUDA_VISIBLE_DEVICES, if any
    cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", "")
    visible_devices = cuda_visible_devices.split(',') if cuda_visible_devices else []
    for device in visible_devices:
        if device not in indices and not any(uuid.startswith(device) for uuid in uuids):
            perror(f"{device} not found")
            return None

    configured, unconfigured = find_in_cdi(visible_devices + ["all"])

    configured_has_all = "all" in configured
    if unconfigured and not configured_has_all:
        perror(f"No CDI configuration found for {','.join(unconfigured)}")
        perror("You can use the \"nvidia-ctk cdi generate\" command from the ")
        perror("nvidia-container-toolkit to generate a CDI configuration.")
        perror("See ramalama-cuda(7).")
        return None
    elif configured:
        if configured_has_all:
            configured.remove("all")
            if not configured:
                configured = indices

        os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(configured)
        return "cuda"

    return None


def check_ascend() -> Optional[Literal["cann"]]:
    try:
        command = ['npu-smi', 'info']
        run_cmd(command, encoding="utf-8")
        os.environ["ASCEND_VISIBLE_DEVICES"] = "0"
        return "cann"
    except Exception:
        pass

    return None


def check_rocm_amd() -> Optional[Literal["hip"]]:
    if is_arm():
        # ROCm is not available for arm64, use Vulkan instead
        return None

    gpu_num = 0
    gpu_bytes = 0
    for i, (np, props) in enumerate(amdkfd.gpus()):
        # Radeon GPUs older than gfx900 are not supported by ROCm (e.g. Polaris)
        if props['gfx_target_version'] < 90000:
            continue

        mem_banks_count = int(props['mem_banks_count'])
        mem_bytes = 0
        for bank in range(mem_banks_count):
            bank_props = amdkfd.parse_props(np + f'/mem_banks/{bank}/properties')
            # See /usr/include/linux/kfd_sysfs.h for possible heap types
            #
            # Count public and private framebuffer memory as VRAM
            if bank_props['heap_type'] in [amdkfd.HEAP_TYPE_FB_PUBLIC, amdkfd.HEAP_TYPE_FB_PRIVATE]:
                mem_bytes += int(bank_props['size_in_bytes'])

        if mem_bytes > MIN_VRAM_BYTES and mem_bytes > gpu_bytes:
            gpu_bytes = mem_bytes
            gpu_num = i

    if gpu_bytes:
        os.environ["HIP_VISIBLE_DEVICES"] = str(gpu_num)
        return "hip"

    return None


def is_arm() -> bool:
    return platform.machine() in ('arm64', 'aarch64')


def check_intel() -> Optional[Literal["intel"]]:
    igpu_num = 0

    if platform.system() == "Windows":
        igpu_num = _check_intel_windows()
    else:
        igpu_num = _check_intel_linux()

    if igpu_num:
        os.environ["INTEL_VISIBLE_DEVICES"] = str(igpu_num)
        return "intel"

    return None


def _check_intel_linux() -> int:
    igpu_num = 0
    # Device IDs for select Intel GPUs.  See: https://dgpu-docs.intel.com/devices/hardware-table.html
    intel_gpus = (
        b"0xe20b",
        b"0xe20c",
        b"0x46a6",
        b"0x46a8",
        b"0x46aa",
        b"0x56a0",
        b"0x56a1",
        b"0x7d51",
        b"0x7dd5",
        b"0x7d55",
    )
    intel_driver_glob_patterns = ["/sys/bus/pci/drivers/i915/*/device", "/sys/bus/pci/drivers/xe/*/device"]
    # Check to see if any of the device ids in intel_gpus are in the device id of the i915 / xe driver
    for fp in sorted([i for p in intel_driver_glob_patterns for i in glob.glob(p)]):
        with open(fp, 'rb') as file:
            content = file.read()
            for gpu_id in intel_gpus:
                if gpu_id in content:
                    igpu_num += 1
    return igpu_num


def _check_intel_windows() -> int:
    """Detect Intel GPUs on Windows via WMI Win32_VideoController."""
    try:
        import wmi  # type: ignore

        w = wmi.WMI()
        return sum(1 for gpu in w.Win32_VideoController() if "intel" in (gpu.Name or "").lower())
    except Exception:
        return 0


def check_mthreads() -> Optional[Literal["musa"]]:
    try:
        command = ['mthreads-gmi']
        run_cmd(command, encoding="utf-8")
        os.environ["MUSA_VISIBLE_DEVICES"] = "0"
        return "musa"
    except Exception:
        pass

    return None


AccelType: TypeAlias = Literal["asahi", "cuda", "cann", "hip", "intel", "musa"]


@lru_cache(maxsize=1)
def get_accel() -> AccelType | Literal["none"]:
    checks: tuple[Callable[[], Optional[AccelType]], ...] = (
        check_asahi,
        cast(Callable[[], Optional[Literal['cuda']]], check_nvidia),
        check_ascend,
        check_rocm_amd,
        check_intel,
        check_mthreads,
    )
    for check in checks:
        if result := check():
            return result
    return "none"


def set_accel_env_vars():
    if get_accel_env_vars():
        return

    get_accel()


def set_gpu_type_env_vars():
    if get_gpu_type_env_vars():
        return

    get_accel()


GPUEnvVar: TypeAlias = Literal[
    "ASAHI_VISIBLE_DEVICES",
    "ASCEND_VISIBLE_DEVICES",
    "CUDA_VISIBLE_DEVICES",
    "GGML_VK_VISIBLE_DEVICES",
    "HIP_VISIBLE_DEVICES",
    "INTEL_VISIBLE_DEVICES",
    "MUSA_VISIBLE_DEVICES",
]


def get_gpu_devices():
    devices = {}
    for dev in ["dri", "kfd", "accel"]:
        path = "/dev/" + dev
        if os.path.exists(path):
            devices[dev] = path
    return dict(sorted(devices.items()))


def get_gpu_type_env_vars() -> dict[GPUEnvVar, str]:
    return {k: v for k in get_args(GPUEnvVar) if (v := os.environ.get(k))}


AccelEnvVar: TypeAlias = Literal[
    "CUDA_LAUNCH_BLOCKING",
    "HSA_VISIBLE_DEVICES",
    "HSA_OVERRIDE_GFX_VERSION",
    "MTHREADS_VISIBLE_DEVICES",
]


def get_accel_env_vars() -> dict[GPUEnvVar | AccelEnvVar, str]:
    gpu_env_vars: dict[GPUEnvVar, str] = get_gpu_type_env_vars()
    accel_env_vars: dict[AccelEnvVar, str] = {k: v for k in get_args(AccelEnvVar) if (v := os.environ.get(k))}
    return gpu_env_vars | accel_env_vars


def rm_until_substring(input: str, substring: str) -> str:
    pos = input.find(substring)
    if pos == -1:
        return input
    return input[pos + len(substring) :]


def minor_release() -> str:
    version_split = version().split(".")
    vers = ".".join(version_split[:2])
    if vers == "0":
        vers = "latest"
    return vers


def version_tagged_image(image: str) -> str:
    if len(image.split(":")) > 1:
        return image
    return f"{image}:{minor_release()}"


def latest_tagged_image(image: str) -> str:
    if len(image.split(":")) > 1:
        return image
    return f"{image}:latest"


class AccelImageArgsWithImage(Protocol):
    image: str


class AccelImageArgsOtherRuntime(Protocol):
    runtime: str
    container: bool
    quiet: bool


class AccelImageArgsOtherRuntimeRAG(Protocol):
    rag: bool
    runtime: str
    container: bool
    quiet: bool


AccelImageArgs: TypeAlias = Union[None, AccelImageArgsOtherRuntime, AccelImageArgsOtherRuntimeRAG]


def accel_image(config: Config, images: Optional[dict[str, str]] = None, conf_key: str = "image") -> str:
    """
    Selects the appropriate image based on config, arguments, environment.
    "images" is a mapping of environment variable names to image names. If not specified,
    the runtime plugin is asked to select the image.
    "conf_key" is the configuration key that holds the configured value of the selected image.
    If not specified, it defaults to "image".

    Never pulls images; callers that need pulling should pass the result to ensure_image().
    """

    # User provided an image via config; tag with :latest if no tag given
    if config.is_set(conf_key):
        return latest_tagged_image(getattr(config, conf_key))

    set_gpu_type_env_vars()
    gpu_type = next(iter(get_gpu_type_env_vars()), "")

    if not images:
        # Ask the runtime plugin to select the image based on detected GPU and its own logic
        from ramalama.plugins.loader import get_runtime

        plugin_image = get_runtime(config.runtime).get_container_image(config, gpu_type)
        if plugin_image is not None:
            return latest_tagged_image(plugin_image)
        images = config.images  # plugin returned None (e.g., MLX); fall back to user dict

    # Explicit images dict provided (e.g., RAG): select by detected GPU type
    return latest_tagged_image(images.get(gpu_type, getattr(config, f"default_{conf_key}")))


def ensure_image(conman: Optional[str], image: str, should_pull: bool = False) -> str:
    """Check if image exists locally; optionally pull it.

    Returns the image string on success. If conman is falsy, returns image unchanged.
    Raises ValueError if the image cannot be pulled.
    """
    if not conman:
        return image

    if ":" not in image:
        image = f"{image}:latest"

    try:
        if run_cmd([conman, "inspect", image], ignore_all=True):
            return image
    except Exception:
        pass

    if not should_pull:
        return image

    try:
        run_cmd([conman, "pull", image], ignore_stderr=True)
        return image
    except Exception:
        pass

    # This is a fallback for main when the version has been bumped but the
    # tagged image for the version has not been pushed yet
    # Only try this for quay.io/ramalama/ images
    base, _, _ = image.rpartition(":")
    if base and base.startswith("quay.io/ramalama/"):
        latest = latest_tagged_image(base)
        try:
            run_cmd([conman, "pull", latest], ignore_stderr=True)
            return latest
        except Exception as e:
            raise ValueError(f"Failed to pull image {image} or {latest}: {e}")

    raise ValueError(f"Failed to pull image {image}")


class ContainerEntryPoint(str):
    def __init__(self, entrypoint: Optional[str] = None):
        self.entrypoint = entrypoint

    def __str__(self):
        return str(self.entrypoint)

    def __repr__(self):
        return repr(self.entrypoint)


SEMVER_RE = re.compile(
    r"^(?P<major>0|[1-9]\d*)\."
    r"(?P<minor>0|[1-9]\d*)\."
    r"(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>"
    r"(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*"
    r"))?"
    r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


def parse_semver(s: str) -> SemVer:
    m = SEMVER_RE.fullmatch(s)
    if not m:
        raise ValueError(f"Not a valid SemVer 2.0.0: {s!r}")
    major = int(m.group("major"))
    minor = int(m.group("minor"))
    patch = int(m.group("patch"))
    return SemVer(major, minor, patch)


@dataclass(frozen=True, order=True)
class SemVer:
    major: int
    minor: int
    patch: int

    @classmethod
    def parse(cls, s: str) -> "SemVer":
        return parse_semver(s)



from __future__ import annotations

import os
import re
import shutil
from pathlib import Path

from ramalama.common import SPLIT_MODEL_PATH_RE, generate_sha256, is_split_file_model
from ramalama.model_store.snapshot_file import SnapshotFile, SnapshotFileType
from ramalama.path_utils import normalize_host_path_for_container
from ramalama.transports.base import Transport
from ramalama.transports.huggingface import HuggingfaceRepository
from ramalama.transports.modelscope import ModelScopeRepository


class LocalModelFile(SnapshotFile):
    def __init__(
        self,
        url,
        header,
        model_file_hash,
        name,
        should_show_progress=False,
        should_verify_checksum=False,
        required=True,
    ):
        super().__init__(
            url,
            header,
            model_file_hash,
            name,
            SnapshotFileType.GGUFModel,
            should_show_progress,
            should_verify_checksum,
            required,
        )

    def download(self, blob_file_path, snapshot_dir):
        if not os.path.exists(self.url):
            raise FileNotFoundError(f"No such file: '{self.url}'")
        # moving from the local location to blob directory so the model store "owns" the data
        shutil.copy(self.url, blob_file_path)
        return os.path.relpath(blob_file_path, start=snapshot_dir)


class URL(Transport):
    def __init__(self, model, model_store_path, scheme):
        self.type = scheme
        super().__init__(model, model_store_path)

        # Use the URL scheme as model type so we can distinguish
        # between the various types such as http, https and file
        self._model_type = scheme

        split = self.model.rsplit("/", 1)
        self.directory = split[0].removeprefix("/") if len(split) > 1 else ""

    @property
    def model_alias(self):
        if self.type == "file":
            return self.model
        return super().model_alias()

    def extract_model_identifiers(self):
        if self.type == "file":
            return (
                Path(self.model).name,
                "latest",
                normalize_host_path_for_container(str(Path(self.model).parent)).removeprefix("/"),
            )

        model_name, model_tag, model_organization = super().extract_model_identifiers()

        parts = model_organization.split("/")
        if len(parts) > 2 and parts[-2] == "blob":
            model_organization = "/".join(parts[:-2])
            model_tag = parts[-1]

        # handling huggingface specific URLs for more precise identifiers
        if len(parts) > 2 and HuggingfaceRepository.REGISTRY_URL.endswith(parts[0]) and parts[-2] == "resolve":
            model_organization = "/".join(parts[:-2])
            model_tag = parts[-1]

        if len(parts) > 3 and parts[-3] == "file":
            model_organization = "/".join(parts[:-3])
            model_tag = parts[-1]

        # handling modelscope specific URLs for more precise identifiers
        if len(parts) > 2 and ModelScopeRepository.REGISTRY_URL.endswith(parts[0]) and parts[-2] == "resolve":
            model_organization = "/".join(parts[:-2])
            model_tag = parts[-1]

        return model_name, model_tag, model_organization

    def _assemble_split_file_list(self, snapshot_hash: str) -> list[SnapshotFile]:
        files: list[SnapshotFile] = []

        # model is split, lets fetch all files based on the name pattern
        match = re.match(SPLIT_MODEL_PATH_RE, self.model)
        if match is None:
            return files

        path_part = match[1]
        if path_part:
            path_part += '/'
        filename_base = match[2]
        total_parts = int(match[3])

        for i in range(1, total_parts + 1):
            file_name = f"{filename_base}-{i:05d}-of-{total_parts:05d}.gguf"
            url = f"{self.type}://{path_part}{file_name}"
            if self.type == "file":
                files.append(
                    LocalModelFile(
                        url=f"{path_part}{file_name}",
                        header={},
                        model_file_hash=generate_sha256(file_name),
                        name=file_name,
                        required=True,
                    )
                )
            else:
                files.append(
                    SnapshotFile(
                        url=url,
                        header={},
                        hash=generate_sha256(file_name),
                        type=SnapshotFileType.GGUFModel,
                        name=file_name,
                        should_show_progress=True,
                        required=True,
                    )
                )

        return files

    def pull(self, args) -> None:
        name, tag, _ = self.extract_model_identifiers()
        _, _, all_files = self.model_store.get_cached_files(tag)
        if all_files:
            return

        files: list[SnapshotFile] = []
        snapshot_hash = generate_sha256(name)
        if is_split_file_model(self.model):
            files = self._assemble_split_file_list(snapshot_hash)
            self.model_store.new_snapshot(tag, snapshot_hash, files, verify=getattr(args, "verify", True))
            return

        if self.type == "file":
            files.append(
                LocalModelFile(
                    url=self.model,  # model contains the full path here
                    header={},
                    model_file_hash=snapshot_hash,
                    name=name,
                    required=True,
                )
            )
            self.model_store.new_snapshot(tag, snapshot_hash, files, verify=getattr(args, "verify", True))
            return

        files.append(
            SnapshotFile(
                url=f"{self.type}://{self.model}",
                header={},
                hash=snapshot_hash,
                type=SnapshotFileType.GGUFModel,
                name=name,
                should_show_progress=True,
                required=True,
            )
        )
        self.model_store.new_snapshot(tag, snapshot_hash, files, verify=getattr(args, "verify", True))
        return



from __future__ import annotations

import argparse
import json
import platform
from collections.abc import Callable
from typing import Optional, cast

from ramalama.arg_types import BaseEngineArgsType
from ramalama.common import run_cmd
from ramalama.config import ActiveConfig
from ramalama.engine import Engine, stop_container
from ramalama.plugins.loader import get_runtime
from ramalama.transports.base import compute_serving_port
from ramalama.transports.transport_factory import New


def _add_common_sandbox_args(parser: argparse.ArgumentParser) -> None:
    """Add --workdir and ARGS arguments shared by all sandbox subcommands."""
    parser.add_argument(
        "-w",
        "--workdir",
        help="local directory to mount into the sandbox container at /work",
    )
    parser.add_argument(
        "ARGS",
        nargs="*",
        help="instructions for the sandbox to process non-interactively",
    )


def add_sandbox_subparsers(subparsers: argparse._SubParsersAction, img_comp: Callable, model_comp: Callable):
    """
    Add subparsers to the provided "subparser" object for each subcommand of
    "ramalama sandbox".
    "img_comp" and "model_comp" are completer functions for images and models, respectively.
    "func" is the function that will be called when the subcommand is run.
    """
    runtime = get_runtime(ActiveConfig().runtime)
    parser = subparsers.add_parser("goose", help="run Goose in a sandbox, backed by a local AI Model")
    if getattr(runtime, "_add_inference_args", None):
        # Consider adding this to the plugin interface for commands which need to run an
        # inference server
        runtime._add_inference_args(parser, "serve")  # type: ignore[attr-defined]
    parser.add_argument("MODEL", completer=model_comp)
    parser.add_argument(
        "--goose-image",
        default="ghcr.io/block/goose:1.28.0",
        completer=img_comp,
        help="Goose container image",
    )
    _add_common_sandbox_args(parser)
    parser.set_defaults(func=run_sandbox_goose)
    yield parser

    parser = subparsers.add_parser("opencode", help="run OpenCode in a sandbox, backed by a local AI Model")
    if getattr(runtime, "_add_inference_args", None):
        runtime._add_inference_args(parser, "serve")  # type: ignore[attr-defined]
    parser.add_argument("MODEL", completer=model_comp)
    parser.add_argument(
        "--opencode-image",
        default="ghcr.io/anomalyco/opencode:1.3.7",
        completer=img_comp,
        help="OpenCode container image",
    )
    _add_common_sandbox_args(parser)
    parser.set_defaults(func=run_sandbox_opencode)
    yield parser


class SandboxEngineArgsType(BaseEngineArgsType):
    ARGS: list[str]
    workdir: Optional[str]


class SandboxEngine(Engine):
    """Engine for running sandbox containers."""

    def __init__(self, args: SandboxEngineArgsType) -> None:
        super().__init__(args)

    def base_args(self) -> None:
        self.add_args("run", "--rm", "-i")

    def is_tty_cmd(self) -> bool:
        return getattr(self.args, "subcommand", "") == "sandbox"

    def add_network(self) -> None:
        self.add_args(f"--network=container:{self.args.name}")  # type: ignore[attr-defined]

    def add_workdir(self, args: SandboxEngineArgsType):
        if args.workdir:
            self.add_volume(args.workdir, "/work", opts="rw")
            self.add_args("--workdir=/work")

    def add_port_option(self) -> None:
        pass

    def add_oci_runtime(self) -> None:
        pass

    def add_detach_option(self) -> None:
        pass

    def add_device_options(self) -> None:
        pass


class GooseArgsType(SandboxEngineArgsType):
    goose_image: str


class Agent:
    """
    Run an agent in a container.
    """

    def __init__(self, args: SandboxEngineArgsType, model_name: str):
        self.engine = SandboxEngine(args)
        self.model_name = model_name

    def run(self) -> None:
        run_cmd(self.engine.exec_args, stdout=None, stdin=None)


class Goose(Agent):
    """
    Run Goose in a sandbox.
    Environment variables required by Goose will be set, and any workdir specified will be mounted into the container.
    If args are provided, they will be passed to Goose to process non-interactively. If there are no arguments and stdin
    is a tty, an interactive session will be started. Otherwise, instructions will be read from stdin.
    """

    def __init__(self, args: GooseArgsType, model_name: str) -> None:
        super().__init__(args, model_name)
        if self.engine.use_podman:
            if platform.system() != "Windows":
                self.engine.add_args("--uidmap=+1000:0")
        self.engine.add_name(f"goose-{args.name}")  # type: ignore[attr-defined]
        self.add_env_options(args)
        self.engine.add_workdir(args)
        self.engine.add_args(args.goose_image)
        if args.ARGS:
            self.engine.add_args("run", "-t", " ".join(args.ARGS))
        elif self.engine.use_tty():
            self.engine.add_args("session")
        else:
            self.engine.add_args("run", "-i", "-")

    def add_env_options(self, args: GooseArgsType) -> None:
        self.engine.add_env_option("GOOSE_PROVIDER=openai")
        self.engine.add_env_option(f"OPENAI_HOST=http://localhost:{args.port}")
        self.engine.add_env_option("OPENAI_API_KEY=ramalama")
        self.engine.add_env_option(f"GOOSE_MODEL={self.model_name}")
        self.engine.add_env_option("GOOSE_TELEMETRY_ENABLED=false")
        self.engine.add_env_option("GOOSE_CLI_SHOW_THINKING=true")


class OpenCodeArgsType(SandboxEngineArgsType):
    opencode_image: str


class OpenCode(Agent):
    """
    Run OpenCode in a sandbox.
    Environment variables required by OpenCode will be set, and any workdir specified will be mounted into the
    container. If args are provided, they will be passed to OpenCode to process non-interactively. If there are
    no arguments and stdin is a tty, an interactive TUI session will be started. Otherwise, instructions will be
    read from stdin.
    """

    def __init__(self, args: OpenCodeArgsType, model_name: str) -> None:
        super().__init__(args, model_name)
        self.engine.add_name(f"opencode-{args.name}")  # type: ignore[attr-defined]
        self.add_env_options(args)
        self.engine.add_workdir(args)
        self.engine.add_args(args.opencode_image)
        if args.ARGS or not self.engine.use_tty():
            # Use the "run" command to process args from the command-line or stdin non-interatively
            self.engine.add_args("run", "--thinking=true")
            self.engine.add(args.ARGS)
        # Running on a tty with no arguments will start the TUI for an interactive session

    def add_env_options(self, args: OpenCodeArgsType) -> None:
        config = {
            "$schema": "https://opencode.ai/config.json",
            "model": f"ramalama/{self.model_name}",
            "provider": {
                "ramalama": {
                    "npm": "@ai-sdk/openai-compatible",
                    "name": "RamaLama",
                    "options": {
                        "baseURL": f"http://localhost:{args.port}/v1",
                        "apiKey": "ramalama",
                    },
                    "models": {
                        self.model_name: {
                            "name": self.model_name,
                        },
                    },
                },
            },
        }
        self.engine.add_env_option(f"OPENCODE_CONFIG_CONTENT={json.dumps(config)}")


def run_sandbox_goose(args: GooseArgsType):
    run_sandbox(args, Goose)


def run_sandbox_opencode(args: OpenCodeArgsType):
    run_sandbox(args, OpenCode)


def run_sandbox(args: SandboxEngineArgsType, agent_cls: type[Agent]):
    """Orchestrate model server and sandbox containers."""

    if not args.container:  # type: ignore[attr-defined]
        raise ValueError("ramalama sandbox requires a container engine")

    args.port = compute_serving_port(args)

    model = New(args.MODEL, args)
    model.ensure_model_exists(args)

    runtime = get_runtime(ActiveConfig().runtime)
    cmd = runtime.handle_subcommand("serve", cast(argparse.Namespace, args))

    model.serve_nonblocking(args, cmd)  # type: ignore[union-attr]

    agent = agent_cls(args, model.model_alias)

    if args.dryrun:
        agent.engine.dryrun()
        return

    try:
        # Wait for model server to be healthy
        model.wait_for_healthy(args)  # type: ignore[union-attr]

        # Launch agent
        agent.run()
    finally:
        args.ignore = True  # type: ignore[attr-defined]
        stop_container(args, args.name, remove=True)  # type: ignore[attr-defined]
